{*************************************************************************}
{*   ZipFormat                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Files\ArchiveFormats\Zip\ZipFormat.pas                       *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 5th November 2003  (05.11.2003)                            }
{*************************************************************************}

{This unit contains class for accessing Zip file archives.}

unit ZipFormat;

{$I 'Compiler.inc'}

interface

uses
  Kernel, Objects, Platforms, Files, Unzip;

type

  {This archive format supports extracting files from Zip archives.}
  TNZIPFormat = class(TNArchiveFormat)
  protected
    procedure DoReadFile(Archive: TNArchive; const FileName: string; var Data:
      Pointer; var Size: LongInt); override;
    function DoOpen(const FileName: string): TNArchive; override;
    procedure DoClose(Archive: TNArchive); override;
    procedure DoFillFileList(Archive: TNArchive); override;
  public
    constructor Create; override;

    function TestFormat(Stream: TNStream): Boolean; override;
  end;

const

  sZIPExtensions = 'zip,pk3,jar,vzf';
  sZIPFormatName = 'Zip Archive';

var

  ZIPInstance: TNZIPFormat = nil;


{*************************************************************************}
{                         IMPLEMEMTATION                                  }
{*************************************************************************}


implementation

uses
  SysUtils;


{*************************************************************************}
{                        TNZIPFormat methods                              }
{*************************************************************************}


constructor TNZIPFormat.Create;
begin
  inherited Create;
  FExtensions := sZIPExtensions;
  FFormatName := sZIPFormatName;
  FCanLoad := True;
  FCanSave := False;
end;

procedure TNZIPFormat.DoReadFile(Archive: TNArchive; const FileName: string; var
  Data: Pointer; var Size: LongInt);
var
  I: Long;
  LocRes: Long;
  Found: Boolean;
  Info: unz_file_info;
  OnlyName: string;
begin
  LocRes := UnzLocateFile(Archive.Handle, PChar(SwapPathDelims(FileName,'/')), 0);
  if (LocRes <> UNZ_OK) then
    if IsFileInArchive(Archive, FileName) then
    begin
      Found := False;
      OnlyName := ExtractFileName(FileName);
      for I := 0 to Archive.Files.Count - 1 do
        if PosNoCase(OnlyName, Archive.Files[I]) > 0 then
          if UnzLocateFile(Archive.Handle, PChar(Archive.Files[I]), 0) <>
            UNZ_OK then
            Continue
          else
          begin
            Found := True;
            Break;
          end;
      if not Found then
        raise Exception.Create('File "' + FileName + '" not found in archive "'
          + Archive.Path + '"!');
    end
    else
      raise Exception.Create('File "' + FileName + '" not found in archive "'
        + Archive.Path + '"!');

  if UnzGetCurrentFileInfo(Archive.Handle, @Info, nil, 0, nil, 0, nil, 0) <>
    UNZ_OK then
    raise Exception.Create('Error when opening file "' + FileName +
      '" in archive "' + Archive.Path + '"!');
  if UnzOpenCurrentFile(Archive.Handle) <> UNZ_OK then
    raise Exception.Create('Error when opening file "' + FileName +
      '" in archive "' + Archive.Path + '"!');

  Size := Info.uncompressed_size;
  GetMem(Data, Size);

  if UnzReadCurrentFile(Archive.Handle, Data, Size) <>
    Long(Info.uncompressed_size) then
    raise Exception.Create('Error when reading file "' + FileName +
      '" in archive "' + Archive.Path + '"!');
  if UnzCloseCurrentFile(Archive.Handle) <> UNZ_OK then
    raise Exception.Create('Error when closing file "' + FileName +
      '" in archive "' + Archive.Path + '"!');
end;

function TNZIPFormat.DoOpen(const FileName: string): TNArchive;
var
  UFile: Pointer;
begin
  Result := nil;
  UFile := Unzip.unzOpen(PChar(FileName));
  if UFile = nil then
    Exit;
  Result := TNArchive.Create(Self, UFile);
  Result.Path := FileName;
  Result.Name := ExtractFileName(FileName);
end;

procedure TNZIPFormat.DoClose(Archive: TNArchive);
begin
  if UnzClose(Archive.Handle) <> UNZ_OK then
     raise Exception.Create('Error when closing ZIP archive!');
end;

procedure TNZIPFormat.DoFillFileList(Archive: TNArchive);
var
  Status: Long;
  FileName: array[0..511] of Char;
begin
  Status := Unzip.UnzGoToFirstFile(Archive.Handle);
  while (Status = UNZ_OK) do
  begin
    if UnzGetCurrentFileInfo(Archive.Handle, nil, FileName, 512, nil, 0,
      nil, 0) = UNZ_OK then
      Archive.Files.Add(FileName);
    Status := UnzGoToNextFile(Archive.Handle);
  end;
end;

function TNZIPFormat.TestFormat(Stream: TNStream): Boolean;
var
  OldPos: Int64;
  Magic: array[0..1] of Char;
begin
  Result := False;
  if Stream = nil then Exit;
  OldPos := Stream.Position;
  if Stream is TNFileStream then
    Result := Unzip.unzlocal_SearchCentralDir(TNFileStream(Stream)) <> 0
  else
  begin
    ZeroMemory(@Magic, SizeOf(Magic));
    Stream.Read(Magic, SizeOf(Magic));
    Result := (Magic[0] = 'P') and (Magic[1] = 'K')
  end;
  Stream.Position := OldPos;
end;


{*************************************************************************}
{                         Internal Stuff                                  }
{*************************************************************************}


initialization

  ZIPInstance := TNZIPFormat.Create;

finalization

  FreeAndNil(ZIPInstance);

end.

