{*************************************************************************}
{*   MemMgrs                                                             *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\MemMgrs.pas                                           *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 22th September 2003  (22.09.2003)                          }
{*************************************************************************}

{Provisional memory manager settings until we have our own
 global crossplatform mem manager.}

unit MemMgrs;

{TODO -oGalfar: Find a way how to add Objects to uses without startup errors
(to replace MemLog with TNLog).}

{$I 'Compiler.inc'}

interface

var
  StdMemoryManager: TMemoryManager;
  NewMemoryManager: TMemoryManager;

  {Memory usage/alloc stats variables}
  GetMemCount: LongInt = 0;
  FreeMemCount: LongInt = 0;
  ReallocMemCount: LongInt = 0;

const
{$IFDEF MSWINDOWS}
  LibName = 'msvcrt.dll';
{$ENDIF}
{$IFDEF LINUX}
  LibName = 'libc.so.6';
{$ENDIF}

function Malloc(Size: LongInt): Pointer;
{$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external LibName name 'malloc';
procedure Free(P: Pointer);
{$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external LibName name 'free';
function ReAlloc(P: Pointer; Size: LongInt): pointer;
{$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external LibName name 'realloc';


{*************************************************************************}
{                         IMPLEMENTATION                                  }
{*************************************************************************}


implementation

uses
  Kernel, SysUtils
{$IFDEF MSWINDOWS}
  , Windows
{$ENDIF};

{$IFDEF MSWINDOWS}
const
  HEAP_NO_SERIALIZE = $00001;
  HEAP_GROWABLE = $00002;
  HEAP_GENERATE_EXCEPTIONS = $00004;
  HEAP_ZERO_MEMORY = $00008;
  HEAP_REALLOC_IN_PLACE_ONLY = $00010;
  HEAP_TAIL_CHECKING_ENABLED = $00020;
  HEAP_FREE_CHECKING_ENABLED = $00040;
  HEAP_DISABLE_COALESCE_ON_FREE = $00080;
  HEAP_CREATE_ALIGN_16 = $10000;
  HEAP_CREATE_ENABLE_TRACING = $20000;
  HEAP_MAXIMUM_TAG = $00FFF;
  HEAP_PSEUDO_TAG_FLAG = $08000;
  HEAP_TAG_SHIFT = 16;

var
  HeapHandle: THandle;
  HeapFlags: ULong = 0;
{$ENDIF}

{$IFDEF DEBUG}
type

  TBlockInfo = record
    Address: Pointer;
    Size: Long;
    Reallocated: Long;
    AllocOrder: Long;
  end;
  PBlockInfo = ^TBlockInfo;

  TBlocks = array[0..MaxLongArraySize div SizeOf(TBlockInfo) - 1] of TBlockInfo;
  PBlocks = ^TBlocks;

  TMsgs = array[0..MaxLongArraySize div 255 - 1] of ShortString;
  PMsgs = ^TMsgs;

var

  Blocks: PBlocks = nil;
  BlockCount: Long = 0;
  Allocations: Long = 0;
  MemLog: TextFile;
  Msgs: PMsgs;
  MsgCount: Long = 0;
  DoDebug: Boolean = True;

function DebugAlloc(Size: Long): Pointer;
begin
{$IFDEF MSWINDOWS}
  Result := HeapAlloc(HeapHandle, HeapFlags, Size);
{$ENDIF}
{$IFDEF LINUX}
  Result := Malloc(Size);
{$ENDIF}
end;

function DebugReAlloc(P: Pointer; Size: Long): Pointer;
begin
  if P <> nil then
{$IFDEF MSWINDOWS}
    Result := HeapRealloc(HeapHandle, HeapFlags and (HEAP_GENERATE_EXCEPTIONS),
     P, Size)
{$ENDIF}
{$IFDEF LINUX}
    Result := ReAlloc(P, Size)
{$ENDIF}
  else
    Result := DebugAlloc(Size);
end;

procedure AddMsg(const Msg: string);
begin
  Inc(MsgCount);
  Msgs := DebugReAlloc(Msgs, MsgCount * SizeOf(ShortString));
  Msgs[MsgCount - 1] := Msg;
end;

procedure RemoveBlock(Idx: Long);
var
  I: Long;
begin
{$IFDEF LOGTOFILES}
  DoDebug := False;
  with Blocks[Idx] do
    AddMsg(
      Format('Pointer freed: Address: %p, Size: %.0n B, AllocOrder: %d',
      [Address, Size / 1, AllocOrder]));
  DoDebug := True;
{$ENDIF}
  for I := Idx to BlockCount - 2 do
    Blocks[I] := Blocks[I + 1];
  Dec(BlockCount);
  Blocks := DebugRealloc(Blocks, BlockCount * SizeOf(TBlockInfo));
end;

procedure AddLB;
begin
  AddMsg('----------------------------------------------------');
end;

procedure ReportOnStartUp;
begin
  DoDebug := False;
  AddLB;
  AddMsg('               Debug Memory Log File');
  AddMsg('                      StartUp       ');
  AddLB;
  DoDebug := True;
end;

procedure ReportOnShutDown;
var
  I: Long;
begin
  AddLB;
  AddMsg('                    ShutDown       ');
  if BlockCount > 0 then
  begin
    AddMsg(Format('There are still %d allocated pointers!',
      [BlockCount]));
    for I := 0 to BlockCount - 1 do
      with Blocks[I] do
        AddMsg(Format('Address: %p, Size: %.0n B, AllocOrder: %d',
          [Address, Size / 1, AllocOrder]));
  end;
  AddLB;
  AddMsg(Format('Pointers allocated: %d', [Kernel.GetMemCount]));
  AddMsg(Format('Pointers reallocated: %d', [Kernel.ReallocMemCount]));
  AddMsg(Format('Pointers freed: %d', [Kernel.FreeMemCount]));
  AddLB;
end;

procedure WriteLog;
var
  I: Long;
begin
  for I := 0 to MsgCount - 1 do
    WriteLn(MemLog, Msgs[I]);
  FreeMem(Msgs, 0);
end;

{$ENDIF}

function NewGetMem(Size: LongInt): Pointer;
begin
{$IFDEF MSWINDOWS}
  Result := HeapAlloc(HeapHandle, HeapFlags, Size);
{$ENDIF}
{$IFDEF LINUX}
  Result := Malloc(Size);
{$ENDIF}

{$IFNDEF DEBUG}
  Inc(GetMemCount);
{$ENDIF}

{$IFDEF DEBUG}
  if DoDebug and (Result <> nil) then
  begin
    Inc(BlockCount);
    Inc(Allocations);
    Inc(GetMemCount);
    Blocks := DebugRealloc(Blocks, BlockCount * SizeOf(TBlockInfo));
    Blocks[BlockCount - 1].Address := Result;
    Blocks[BlockCount - 1].Size := Size;
    Blocks[BlockCount - 1].AllocOrder := Allocations;
    Blocks[BlockCount - 1].Reallocated := 0;
{$IFDEF LOGTOFILES}
    DoDebug := False;
    with Blocks[BlockCount - 1] do
      AddMsg(
        Format('Pointer allocated: Address: %p, Size: %.0n B, AllocOrder: %d',
        [Address, Size / 1, AllocOrder]));
    DoDebug := True;
{$ENDIF}
  end;
{$ENDIF}
end;

function NewFreeMem(P: Pointer): LongInt;
{$IFDEF DEBUG}
var
  I: Long;
{$ENDIF}
begin
  Result := -1;
  if P <> nil then
  begin
{$IFDEF MSWINDOWS}
    Result := Long(not HeapFree(HeapHandle, HeapFlags,P));
{$ENDIF}
{$IFDEF LINUX}
    Free(P);
    Result := 0;
{$ENDIF}

{$IFNDEF DEBUG}
    Inc(FreeMemCount);
{$ENDIF}

{$IFDEF DEBUG}
    if DoDebug then
      for I := 0 to BlockCount - 1 do
        if Blocks[I].Address = P then
        begin
          Inc(FreeMemCount);
          RemoveBlock(I);
          Break;
        end;
{$ENDIF}
  end;
end;

function NewReallocMem(P: Pointer; Size: LongInt): Pointer;
{$IFDEF DEBUG}
var
  I: Long;
{$ENDIF}
begin
  if P <> nil then
  begin
{$IFDEF MSWINDOWS}
    Result := HeapRealloc(HeapHandle, HeapFlags and (HEAP_GENERATE_EXCEPTIONS),
     P, Size);
{$ENDIF}
{$IFDEF LINUX}
    Result := ReAlloc(P, Size);
{$ENDIF}

{$IFNDEF DEBUG}
    Inc(ReallocMemCount);
{$ENDIF}

{$IFDEF DEBUG}
    if DoDebug then
      for I := 0 to BlockCount - 1 do
        if Blocks[I].Address = P then
        begin
          Inc(ReallocMemCount);
          if Result <> nil then
          begin
            Blocks[I].Size := Size;
            Inc(Blocks[I].Reallocated);
{$IFDEF LOGTOFILES}
            DoDebug := False;
            with Blocks[I] do
              AddMsg(
                Format('Pointer reallocated: Address: %p, Size: %.0n B,' +
                ' AllocOrder: %d, Reallocated: %d',
                [Address, Size / 1, AllocOrder, Reallocated]));
            DoDebug := True;
{$ENDIF}
          end
          else
            RemoveBlock(I);
          Break;
        end;
{$ENDIF}
  end
  else
    GetMem(Result, Size);
end;

procedure SetMemMgr;
begin
  NewMemoryManager.GetMem := NewGetMem;
  NewMemoryManager.FreeMem := NewFreeMem;
  NewMemoryManager.ReallocMem := NewReallocMem;
  GetMemoryManager(StdMemoryManager);
  SetMemoryManager(NewMemoryManager);
end;

procedure UnSetMemMgr;
begin
  SetMemoryManager(StdMemoryManager);
end;


initialization

{$IFDEF MSWINDOWS}
  HeapHandle := GetProcessHeap;
{$ENDIF}
  SetMemMgr;
{$IFDEF DEBUG}
{$IFDEF LOGTOFILES}
  ReportOnStartUp;
{$IFNDEF MODLIBRARY}
  try
    ForceDirectories(ExeDir + LogDir);
    AssignFile(MemLog, ExeDir + LogDir + PathDelim + ExtractFileName(
      ChangeFileExt(ExeName, '.Memory.log')));
    Rewrite(MemLog);
  except
    // do not show anything
  end;
{$ENDIF}
{$ENDIF}
  DoDebug := True;
{$ENDIF}

finalization

{$IFDEF DEBUG}
  DoDebug := False;
{$IFDEF LOGTOFILES}
  ReportOnShutDown;
{$IFNDEF MODLIBRARY}
  try
    WriteLog;
    CloseFile(MemLog);
  except
    // do not show anything
  end;
{$ENDIF}  
{$ENDIF}
{$ENDIF}
  UnSetMemMgr;

end.

