{*************************************************************************}
{*   Objects                                                             *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Objects.pas                                           *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12nd August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains base, moreless abstract classes declarations.
 There are three types of classes in Numalian Engine:
  1) Singletons - there is always one instance of this class. Used mainly
      as plugins, engines, ...
  2) Stored - like TPersistent's descendants in VCL. Their published properties
      can be saved. Useful for GUI forms, Images, other  data objects ...
  3) Other - like lists, streams and other.

 There are also redefinitions of some std Delphi's classes with our names.
 You also do not need to add 'Classes, Objects' but only 'Objects' to uses.}

unit Objects;

{DONE -oGalfar : Add TNList and other which will not raise excepts when
index is out of bounds.}
{DONE -Galfar : Look at XMLStreaming.pas from FPC. It is exactly what I
want for Stored objects}
{TODO -Galfar : Do not forget to use message handler procedures}
{DONE -Galfar : When lib loading is done add - Classes.UnloadModuleClasses}
{DONE -Galfar : Hashmap or stringlist for storing registered classes.}
{DONE -Galfar : Add (nice) Log file support.}
{DONE -Galfar : Add support for loading module DLLs.
Distinguish by XXX to IFF(imagefileformat),... For
all modules of each type call some registered callback function
(e.g. In the case of IFF, calback will register TNImageFileFormat instance.)}
{TODO -Galfar : Try creating fake CDROM in NERO and put Exe
on it. Add some donotstoresetttingsandlogs option for this case.}
{TODO -oGalfar : Add streaming support for persistents adopted from Classes
(Strings,...) ?how?multipleinheritance needed.}
{DONE -oGalfar: Add better extension checking for file modules -
  Find('z') may return zlib or lzh, ...}

{$I 'Compiler.inc'}

interface

uses
  Classes, SysUtils, Kernel, TypInfo;


{*************************************************************************}
{                         Types/Constants                                 }
{*************************************************************************}


const

  {Some usefull strings.}
  sExceptMsg = 'Exception Message: ';


{*************************************************************************}
{                         Base Classes                                    }
{*************************************************************************}


type

  {Basic class from which all other shalt arise.
   It also supports interfaces.}
  TNBaseObject = class(TInterfacedObject)
  protected
    FID: ULong;

    function GetID: ULong; virtual;
    {Each object can receive messages and this procedure
     can process them. You can use message methods
     (procedure Umajo(var Msg);message KM_Umajo;)
     as used in VCL but it is needless to write new procs for short message
     handlers - so put them in one (this) procedure.
     Do not call directly !! - use Dispatch method always.
     Be careful with sending messages like KM_Free, ...}
    procedure Perform(var Message: TMsgRec); virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    {Overrides basic TObject's message dispatching method so
     it calls Perform or associated message methods.}
    procedure DefaultHandler(var Message); override;

    {Objects Unique identifier.}
    property ID: ULong read GetID;
  end;

  {Class refernce for TNBaseObject.}
  TNBaseObjectClass = class of TNBaseObject;

  {Base class for all singletons in engine.}
  TNSingleton = class(TNBaseObject)
  protected
    FName: string;
    FDescription: string;

    function GetName: string; virtual;
    function GetDescription: string; virtual;

    procedure SetName(const Value: string); virtual;
    procedure SetDescription(const Value: string); virtual;
  public
    constructor Create; override;
    destructor Destroy; override;

    {Singleton's name. Should be unique in descendants.}
    property Name: string read GetName write SetName;
    {Short description of singleton's function.}
    property Description: string read GetDescription write SetDescription;
  end;

  {Class refernce for TNSingleton.}
  TNSingletonClass = class of TNSingleton;

  TNFiler = class;
  TNAbstractReader = class;
  TNAbstractWriter = class;

{$M+} // Runtime type info: on

  {Base class for all 'stored' objects. Published members
   of these classes can be streamed similarly as TPersistent
   descendants in VCL.}
  TNPersistent = class(TNBaseObject)
  protected
    {Assigns self to destination persistent object.}
    procedure AssignTo(Dest: TNPersistent); virtual;
  public
    {Assigns from source persistent object}
    procedure Assign(Source: TNPersistent); virtual;
    {Defines new property for streaming. Use
     Filer.DefineProperty() in overrided methods.}
    procedure DefineProperties(Filer: TNFiler); virtual;
  end;

{$M-} // Runtime type info: off

  {Class refernce for TNPersistent.}
  TNPersistentClass = class of TNPersistent;

  {Persistent with Name.}
  TNNamedPersistent = class(TNPersistent)
  protected
    FName: string;

    procedure SetName(const Value: string); virtual;
  published
    property Name: string read FName write SetName;
  end;


{*************************************************************************}
{                            Lists                                        }
{*************************************************************************}


  {Just to keep 'N' in class names.}
  TNList = Classes.TList;

  {Ordered list from ContNRS.pas - I do not want to add this unit
   to project. Some new methods were added too.}
  TNOrderedList = class(TObject)
  private
    FList: TNList;
  protected
    procedure PushItem(AItem: Pointer); virtual; abstract;
    function PopItem: Pointer; virtual;
    function PeekItem: Pointer; virtual;
  public
    constructor Create;
    destructor Destroy; override;

    function Count: Long;
    function AtLeast(ACount: Long): Boolean;
    function Push(AItem: Pointer): Pointer;
    function Pop: Pointer;
    function Peek: Pointer;
    procedure Clear;
  end;

  {Stack implementation.}
  TNStack = class(TNOrderedList)
  protected
    procedure PushItem(AItem: Pointer); override;
  end;

  {Stack implementation.}
  TNQueue = class(TNOrderedList)
  protected
    procedure PushItem(AItem: Pointer); override;
  end;

  {Just to keep 'N' in class names.}
  TNStrings = Classes.TStrings;

  {This is std Borland's stringlist with much faster comparsions.}
  TNStringList = class(Classes.TStringList)
  protected
    function CompareStrings(const S1, S2: string): Long; override;
  end;

  {This is same as TStringList only Sorted property is set to true in creation
   to enable Find method without additional code writing =
   instead of "S:=TStringlist.Create;S.Sorted:=true" , but S:=TSortedStrings.Create
   It is mainly for lazy persons (like me)}
  TNSortedStrings = class(TNStringList)
  public
    constructor Create;
  end;


{*************************************************************************}
{                         Streaming support                               }
{*************************************************************************}


  {Just to keep 'N' in class names.}
  TNStream = Classes.TStream;
  TNCustomMemoryStream = Classes.TCustomMemoryStream;
  TNMemoryStream = Classes.TMemoryStream;
  TNHandleStream = Classes.THandleStream;
  TNFileStream = Classes.TFileStream;
  TNResourceStream = Classes.TResourceStream;
  TNStringStream = Classes.TStringStream;

  {Procedure used by TNFilers to read custom properties.}
  TReadDataProc = procedure(Reader: TNAbstractReader) of object;
  {Procedure used by TNFilers to write custom properties.}
  TWriteDataProc = procedure(Writer: TNAbstractWriter) of object;

  {Base class for streaming persistent objects and other variables.
   Descendants are TNReader and TNWriter in Streaming.pas.}
  TNFiler = class(TNBaseObject)
  protected
    FStream: TNStream;
    FRoot: TNPersistent;
    FLocalRoot: TNPersistent;
    FRootStack: TNStack;

    procedure SetRoot(Value: TNPersistent);
  public
    constructor Create(AStream: TStream); reintroduce; virtual;
    destructor Destroy; override;

    {This method updates filer's buffer or stream.
     In TNWriter it writes its buffer to stream. In TNReader reads
     buffer from stream. Call this method before/after reading/writing
     to be sure that buffers and streams are up to date.}
    procedure Flush; virtual; abstract;

    {Defines new property to be read/written when streaming persistent
     objects. Look at overrided methods in TNReader and TNWriter.
     It is called by TNPersistent.DefineProperties (in descendants).}
    procedure DefineProperty(const Name: string; ReadData: TReadDataProc;
      WriteData: TWriteDataProc); virtual; abstract;

    {Used when loading nested objects from stream.}
    property Root: TNPersistent read FRoot write SetRoot;
    {Used when loading nested objects from stream.}
    property LocalRoot: TNPersistent read FLocalRoot;
  end;

  {Pure abstract class which contains only public methods of TNReaders.
   It exists because of TReadDataProc which must take filer
   with reading capabilities as prameter but it can not be
   directly TNReader because it is in another unit.}
  TNAbstractReader = class(TNFiler)
  public
    {Returns true if next value in stream is list of values of
     given name. If so begins reading of list. Must be accompanied
     with EndList() call}
    function BeginList(const Name: string): Boolean; virtual; abstract;
    {Returns true if end of list is next and ends reading of this list.}
    function EndList: Boolean; virtual; abstract;
    {ReadXXX functions read data from the form
     specific for given TNReader descendant.}
    function ReadString: string; virtual; abstract;
    function ReadIdent: string; virtual; abstract;
    function ReadChar: Char; virtual; abstract;
    function ReadLong: Long; virtual; abstract;
    function ReadULong: ULong; virtual; abstract;
    function ReadInt64: Int64; virtual; abstract;
    function ReadBool: Boolean; virtual; abstract;
    function ReadSingle: Single; virtual; abstract;
    function ReadDouble: Double; virtual; abstract;
    {Returns nil when fails, otherwise returns newly allocated pointer}
    procedure ReadBinary(var Data: Pointer; var Count: Long); virtual; abstract;
    {Reads persistent object with all its subobjects and properties.
     Returns nil when fails, otherwise returns newly created instance of
     read class.}
    procedure ReadInstance(var Instance: TNPersistent); virtual; abstract;
  end;

  {Pure abstract class which contains only public methods of TNWriters.
   It exists because of TWriteDataProc which must take filer
   with writing capabilities as prameter but it can not be
   directly TNWriter because it is in another unit.}
  TNAbstractWriter = class(TNFiler)
  public
    {Begins new list of values. Next calls to WriteXXX will
     add data to active list (if any). E.g. in TNXMLWriter
     it creates new node. List must be ended with EndList() call.}
    procedure BeginList(const Name: string); virtual; abstract;
    {Ends active list.}
    procedure EndList; virtual; abstract;
    {WriteXXX procedures write their parameters in the form
     specific for given TNWriter descendant. But it always
     writes also info about parameter's type or size.}
    procedure WriteString(const Value: string); virtual; abstract;
    procedure WriteIdent(const Value: string); virtual; abstract;
    procedure WriteChar(const Value: Char); virtual; abstract;
    procedure WriteInt(const Value: Long); overload; virtual; abstract;
    procedure WriteInt(const Value: ULong); overload; virtual; abstract;
    procedure WriteInt(const Value: Int64); overload; virtual; abstract;
    procedure WriteBool(const Value: Boolean); virtual; abstract;
    procedure WriteFloat(const Value: Single); overload; virtual; abstract;
    procedure WriteFloat(const Value: Double); overload; virtual; abstract;
    procedure WriteBinary(Data: Pointer; Count: Long); virtual; abstract;
    {Wrtites object to stream with all its subobjects and properties.}
    procedure WriteInstance(Instance: TNPersistent); virtual; abstract;
  end;

const

  {Stream Types/Consts}
  fmCreate = Classes.fmCreate;

  soFromBeginning = Classes.soFromBeginning;
  soFromCurrent = Classes.soFromCurrent;
  soFromEnd = Classes.soFromEnd;


{*************************************************************************}
{                          Other classes                                  }
{*************************************************************************}


type

  {In some unit this were only things used from SysUtils. Until now...}
  Exception = SysUtils.Exception;
  ExceptClass = SysUtils.ExceptClass;

const

  {Max number of lines in TNLog's message buffer.}
  MaxLogBuffer = {$IFNDEF DEBUG}32{$ELSE}1{$ENDIF};
  {Max number of characters per line. Messages are wrapped after this position.}
  MaxLogLine = 80;
  {Extension of TNLog file.}
  sLogFileExt = 'log';
  {Char used to create linebreaks in log file.}
  sLogLBChar = '-';
  {Log file Ident increase/decrease.}
  sLogIdent = #32#32;

type

  {Class for buffered storing messages during program flow.
   Supports nested and repeated messages handling. Wraps messages
   when they exceed FLineLength length.
   It adds messages to FBuffer and when it is 'full' writes messages
   to file (with name AFileName+sLogFileExt).}
  TNLog = class(TNSingleton)
  protected
    FBuffer: TNStringList;
    FBufMax: Long;
    FStream: TNFileStream;
    FIdent: string;
    FLastMsg: string;
    FStartTime: Int64;
    FRepeats: Long;
    FCheckRepeats: Boolean;
    FWriteToFile: Boolean;
    FLineLength: Long;
    FFileName: string;

    procedure AddStartup; virtual;
    procedure AddShutdown; virtual;
    procedure IncIdent;
    procedure DecIdent;
  public
    {Creates new log file. If file can not be created logging is done
     into memory (stringlist). FileName extension is always overrided
     by sLogExt constant.}
    constructor Create(const AFileName: string); reintroduce; virtual;
    destructor Destroy; override;

    {Writes buffer to file and clears it.}
    procedure Flush;

    {Adds simple string message. When buffer is full calls Flush().}
    procedure Add(const Msg: string); overload;
    {Adds formatted string message.}
    procedure Add(const MsgFmt: string; Args: array of const); overload;
    {Adds message with current time and microseconds from program startup.}
    procedure AddTime(const Msg: string);
    {Adds error message.}
    procedure AddError(const Msg: string);
    {Adds warning message.}
    procedure AddWarning(const Msg: string);
    {Adds Msg then increases Ident. Accompany it with EndSubList() call.}
    procedure BeginSubList(const Msg: string);
    {Decreases Ident and adds Msg.}
    procedure EndSubList(const Msg: string);
    {Adds (80-Length(Ident)) long line break of sLBChar.}
    procedure AddLineBreak;

    {Current Ident.}
    property Ident: string read FIdent;
    {If True repeating messages are written only once with number of
     repeats added.}
    property CheckRepeats: Boolean read FCheckRepeats write FCheckRepeats;
    {Name of file to which this log instance is assigned. Set by constructor.}
    property FileName: string read FFileName;
  end;

  {Descendant of TNLog. Some stratup and shutdown info is added.
   Accessible trough Log() function. Put Kernel related messages here.}
  TNKernelLog = class(TNLog)
  protected
    procedure AddStartup; override;
    procedure AddShutdown; override;
  end;


{*************************************************************************}
{                      Plugin/Module Support                              }
{*************************************************************************}


const

  {Name prefix for module libraries. E.g. ModDirect3D.ren, ModGif.iff,...}
  sModulePrefix = 'Mod';
  {Name of TGetIDProc in module library.}
  sGetIDProc = 'GetID';
  {Name of TGetCountProc in module library.}
  sGetCountProc = 'GetCount';
  {Name of TGetModuleProc in module library.}
  sGetModuleProc = 'GetModule';

type

  {Module identifier. Used to determine which type of module is
   in library. Actuall values are defined in units which
   declare some new modules. E.g. in Images.pas it is IffLibraryID = 'iff'.
   You can change library file extension to  its ID to be more transparent.}
  TModuleID = array[0..2] of Char;

  {(*}

  {Record with information about module library. These records are
   stored by TModuleRegister.}
  TModuleEntry = record
    Name: string;             // Name of module e.g. ModMng.iff
    FileName: string;         // Full path to module e.g. D:\project\ModMng.iff
    Handle: THandle;          // Handle to loaded library
    ID: TModuleID;            // Module identifier
    Count: Long;              // Number of unique modules in library
  end;
  PModuleEntry = ^TModuleEntry;

  {*)}

  {This procedure must be in all valid module libraries. It returns
   identifier for contained modules.}
  TGetIDProc = procedure(var ID: TModuleID); cdecl;

  {This procedure must be in all valid module libraries. It returns
   number of unique modules in library.}
  TGetCountProc = procedure(var Count: Long); cdecl;

  {This procedure must be in all valid module libraries. It returns
   pointer to module identified by Index (0 .. GetCountProc-1).
   It can point to function, class, record, ... }
  TGetModuleProc = procedure(Index: Long; var Module: Pointer); cdecl;

  {This callback (list of them stored in TNModuleRegister) is
   called when new module is loaded form library. Based on
   module's ID callback performs some things (e.g. registers
   new ImageFileFormat so it can be used by image loading/saving).}
  TModLoadCallBack = procedure(Entry: PModuleEntry; var Done: Boolean);
  {This callback (list of them stored in TNModuleRegister) is
   called when module (from library) is unloaded. Based on
   module's ID callback performs some things (e.g. unregisters
   new ImageFileFormat so it can NOT be used by image loading/saving any more).}
  TModUnLoadCallBack = procedure(Entry: PModuleEntry; var Done: Boolean);

  {Base class for modules.}
  TNModule = class(TNBaseObject)
  end;

  {Class reference for TNModule.}
  TNModuleClass = class of TNModule;

  {Base class for file formats loading/saving modules.}
  TNFileModule = class(TNModule)
  protected
    FExtensions: string;
    FFormatName: string;
    FCanLoad: Boolean;
    FCanSave: Boolean;
  public
    {This function tests if input data are in supported format. It must then
     set stream's position to original pos.}
    function TestFormat(Stream: TNStream): Boolean; virtual;

    {File name extensions used by this format.
     Separate them by semicolon or comma.
     Example:= 'jpg,jpeg,jfif'}
    property Extensions: string read FExtensions;
    {Name of file format (e.g. 'Portable Network Graphics').}
    property FormatName: string read FFormatName;
    {Indicates that class can load files.}
    property CanLoad: Boolean read FCanLoad;
    {Indicates that class can save files.}
    property CanSave: Boolean read FCanSave;
  end;

  {List for storing file module instances. For example in Images.pas is
   TNImageFileFormats to hold registered image file formats.
   They are mostly declared in implementation part of units}
  TNFileModuleList = class(TNBaseObject)
  protected
    FModules: TNList;
    FDefault: TNFileModule;

    function GetDefault: TNFileModule;
  public
    constructor Create; override;
    destructor Destroy; override;

    {Adds new module to list.}
    procedure Add(Module: TNFileModule);
    {Removes module to list.}
    procedure Remove(Module: TNFileModule);
    {Returns True if given module is in list.}
    function Find(Module: TNFileModule): Boolean; overload;
    {Finds module which supports given file extension. If not
     fount nil is returned.}
    function Find(const Ext: string): TNFileModule; overload;

    {List of all registered modules.}
    property Modules: TNList read FModules;
    {Default module - by default it is set to module at index 0
     can can be changed as you want. Defaults are for example
     TGA file format for images and ZIP archive format fo archives.}
    property Default: TNFileModule read GetDefault write FDefault;
  end;

{$IFNDEF MODLIBRARY}

  {This class manages modules from external libraries. It stores
   TModuleEntry record for each loaded module and calls
   callback functions when module is loaded/unloaded.}
  TNModuleRegister = class(TNBaseObject)
  protected
    FList: TNSortedStrings;
    FLoadCalls: TNList;
    FUnLoadCalls: TNList;
    FDefaultDirectory: string;

    {Unloads all modules.}
    procedure UnLoadAll;
    {Runs all registered module load callbacks.}
    procedure RunLoadCallBacks(Entry: PModuleEntry);
    {Runs all registered module unload callbacks.}
    procedure RunUnLoadCallBacks(Entry: PModuleEntry);
  public
    constructor Create; override;
    destructor Destroy; override;

    {Loads new module from FileName and runs load callbacks.}
    procedure Load(const FileName: string);
    {Loads all modules from Dir.}
    procedure LoadFromDirectory(const Dir: string);
    {UnLoads module with given FileName and runs unload callbacks.}
    procedure UnLoad(const FileName: string);

    {Adds new load callback.}
    procedure AddLoadCallBack(Call: TModLoadCallBack);
    {Adds new unload callback.}
    procedure AddUnLoadCallBack(Call: TModUnLoadCallBack);

    {Returns True if given module is loaded.}
    function IsLoaded(const FileName: string): Boolean;
    {Returns True if given module is valid module library.}
    function IsValidModuleLibrary(const FileName: string): Boolean;

    {Default directory where ModuleRegister looks for
     modules with non-fullpath FileName.}
    property DefaultDirectory: string read FDefaultDirectory;
  end;

{$ENDIF}


{*************************************************************************}
{                         Class/Object functions                          }
{*************************************************************************}


{This function generates new IDs. It is very simple but why complicate it?.
 However it will repeat numbers after creation of 4294k objects.}
function GetNewID: ULong;

{Number of created TNBaseObject descendants from engine startup.}
function GetCreatedObjects: Long;
{Number of free TNBaseObject descendants from engine startup.}
function GetFreedObjects: Long;
{Number of actually "living" TNBaseObject descendants.}
function GetLivingObjects: Long;

{Registers class so it can be used by streaming system.
 Supports classes with same name from different units.}
procedure RegisterClass(AClass: TNPersistentClass);
{Registers multiple classes at once.}
procedure RegisterClasses(AClasses: array of TNPersistentClass);
{Unregisters registered class.
 Supports classes with same name from different units.}
procedure UnRegisterClass(AClass: TNPersistentClass);
{Unegisters multiple classes at once.}
procedure UnRegisterClasses(AClasses: array of TNPersistentClass);
{Returns registered class reference of given classname.
 If not found it returns nil.Supports classes with same name from different
 units.}
function GetClass(AClassName: string): TNPersistentClass;


{*************************************************************************}
{                           Error functions                               }
{*************************************************************************}


{Use this procedure as exception handler in try...except clauses.
 It shows error message with continue/terminate app option and logs
 error message.}
procedure ShowError(const Message: string; const ExClassName: string = '');


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


{This function provides access to TNLog singleton instance.
 I rather use function than global variable becaude it can check
 if instance is valid.}
function Log: TNLog;

{$IFNDEF MODLIBRARY}

{This function provides access to TNModuleRegister singleton instance.
 I rather use function than global variable becaude it can check
 if instance is valid.}
function ModuleRegister: TNModuleRegister;

{$ENDIF}


{*************************************************************************}
{                         IMPLEMEMTATION                                  }
{*************************************************************************}


implementation

uses
  Platforms;

var

  {Object statistics}
  NowID: ULong = 1;
  ObjCreated: Long = 0;
  ObjFreed: Long = 0;

  {List constaining registered classes.
   Each class used by streaming system must be first registered by
   RegisterClass() call.}
  RegClasses: TNSortedStrings = nil;

  {Global instance of Log file accessible trough Log() function
   (it check if instance is created).}
  LogInstance: TNLog = nil;

{$IFNDEF MODLIBRARY}

  {Global instance of ModuleRegister accessible trough ModuleRegister() function
   (it check if instance is created).}
  ModRegInstance: TNModuleRegister = nil;

{$ENDIF}


{*************************************************************************}
{                         Class/Object functions                          }
{*************************************************************************}


function GetNewID: ULong;
begin
  Result := NowID;
  Inc(NowID);
end;

function GetCreatedObjects: Long;
begin
  Result := ObjCreated;
end;

function GetFreedObjects: Long;
begin
  Result := ObjFreed;
end;

function GetLivingObjects: Long;
begin
  Result := Abs(ObjCreated - ObjFreed);
end;

function GetRegClasses: TNSortedStrings;
begin
  if RegClasses = nil then RegClasses := TNSortedStrings.Create;
  Result := RegClasses;
end;

procedure RegisterClass(AClass: TNPersistentClass);
var
  Name: string;
begin
  if AClass = nil then Exit;
  Name := GetTypeData(PTypeInfo(AClass.ClassInfo)).UnitName + '.' +
    AClass.ClassName;
  if GetRegClasses.AddObject(Name, TObject(AClass)) > -1 then
    Log.Add('Class registered: ' + Name);
end;

procedure RegisterClasses(AClasses: array of TNPersistentClass);
var
  I: Long;
begin
  for I := Low(AClasses) to High(AClasses) do
    RegisterClass(AClasses[I]);
end;

procedure UnRegisterClass(AClass: TNPersistentClass);
var
  FullName: string;
  I: Long;
begin
  if AClass = nil then Exit;
  FullName := GetTypeData(PTypeInfo(AClass.ClassInfo)).UnitName + '.' +
    AClass.ClassName;
  if GetRegClasses.Find(FullName, I) then
  begin
    GetRegClasses.Delete(I);
    Log.Add('Class unregistered: ' + FullName);
  end;
end;

procedure UnRegisterClasses(AClasses: array of TNPersistentClass);
var
  I: Long;
begin
  for I := Low(AClasses) to High(AClasses) do
    RegisterClass(AClasses[I]);
end;

function GetClass(AClassName: string): TNPersistentClass;
var
  I, P: Long;
  S: string;
begin
  // first check incomming classname and if not found
  // check classname without its unit name
  Result := nil;
  if AClassName = '' then Exit;
  I := -1;
  if not GetRegClasses.Find(AClassName, I) then
  begin
    P := Pos('.', AClassName);
    if P > 0 then
    begin
      S := Copy(AClassName, P + 1, Length(AClassName) - P);
      if not GetRegClasses.Find(AClassName, I) then
        I := -1;
    end;
  end;
  if I >= 0 then Result := TNPersistentClass(GetRegClasses.Objects[I]);
end;


{*************************************************************************}
{                           Error functions                               }
{*************************************************************************}


procedure ShowError(const Message: string; const ExClassName: string);
var
  P: Long;
  Msg, EMsg: string;
begin
  Msg := Message;
  // checks if err msg was created by Kernel.ErrorStr
  P := Pos('Message:', Msg);
  EMsg := Msg;
  if P > 0 then
    // delete long error message header
    Delete(EMsg, 1, P + 9);
  // replace all line breaks with spaces
  EMsg := StrReplace(EMsg, sLineBreak, #32);
  Log.AddError(EMsg);
  // insert exception class name
  if (P > 0) and (ExClassName <> '') then
    Insert('Exception Class: ' + ExClassName + sLineBreak, Msg, P);
  case Platform.ErrorMsg(Msg) of
    emTerminate: Halt(1);
    emContinue: ;
  end;
end;


{*************************************************************************}
{                           Singleton access                              }
{*************************************************************************}


function Log: TNLog;
begin
  if LogInstance = nil then LogInstance := TNKernelLog.Create(ExeName);
  Result := LogInstance;
end;

{$IFNDEF MODLIBRARY}

function ModuleRegister: TNModuleRegister;
begin
  if ModRegInstance = nil then ModRegInstance := TNModuleRegister.Create;
  Result := ModRegInstance;
end;

{$ENDIF}

{*************************************************************************}
{                         TNBaseObject Methods                            }
{*************************************************************************}


constructor TNBaseObject.Create;
begin
  FID := GetNewID;
  Inc(ObjCreated);
end;

destructor TNBaseObject.Destroy;
begin
  Inc(ObjFreed);
  inherited Destroy;
end;

function TNBaseObject.GetID: ULong;
begin
  Result := FID;
end;

procedure TNBaseObject.DefaultHandler(var Message);
begin
  inherited DefaultHandler(Message);
  Perform(TMsgRec(Message));
end;

procedure TNBaseObject.Perform(var Message: TMsgRec);
begin
  case Message.MsgID of
    KM_GetID: Message.LParam := ID;
    KM_FreeSelf: Self.Free;
    KM_FreeSender: Message.Sender.Free;
    else
    // in descendants put here this line
    // inherited Perform(..);
    // to perform messages not processed by this handler
  end;
end;


{*************************************************************************}
{                         TNSingleton methods                             }
{*************************************************************************}


constructor TNSingleton.Create;
begin
  inherited Create;
  FName := 'Unknown Singleton';
  FDescription := 'Singleton with unknown capabilities';
end;

destructor TNSingleton.Destroy;
begin
  inherited Destroy;
end;

function TNSingleton.GetName: string;
begin
  Result := FName;
end;

function TNSingleton.GetDescription: string;
begin
  Result := FDescription;
end;

procedure TNSingleton.SetName(const Value: string);
begin
  FName := Value;
end;

procedure TNSingleton.SetDescription(const Value: string);
begin
  FDescription := Value;
end;


{*************************************************************************}
{                             TNPersistent methods                        }
{*************************************************************************}


procedure TNPersistent.AssignTo(Dest: TNPersistent);
begin
end;

procedure TNPersistent.Assign(Source: TNPersistent);
begin
  if Source <> nil then
    Source.AssignTo(Self);
end;

procedure TNPersistent.DefineProperties(Filer: TNFiler);
begin
end;


{*************************************************************************}
{                             TNPersistent methods                        }
{*************************************************************************}


procedure TNNamedPersistent.SetName(const Value: string);
begin
  FName := Value;
end;


{*************************************************************************}
{                          TNOrderedList methods                          }
{*************************************************************************}


constructor TNOrderedList.Create;
begin
  inherited Create;
  FList := TNList.Create;
end;

destructor TNOrderedList.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

function TNOrderedList.PeekItem: Pointer;
begin
  Result := FList[FList.Count - 1];
end;

function TNOrderedList.PopItem: Pointer;
begin
  Result := PeekItem;
  FList.Delete(FList.Count - 1);
end;

function TNOrderedList.AtLeast(ACount: Long): Boolean;
begin
  Result := FList.Count >= ACount;
end;

function TNOrderedList.Peek: Pointer;
begin
  Result := PeekItem;
end;

function TNOrderedList.Pop: Pointer;
begin
  Result := PopItem;
end;

function TNOrderedList.Push(AItem: Pointer): Pointer;
begin
  PushItem(AItem);
  Result := AItem;
end;

function TNOrderedList.Count: Long;
begin
  Result := FList.Count;
end;

procedure TNOrderedList.Clear;
begin
  FList.Clear;
end;


{*************************************************************************}
{                         TNStack methods                                 }
{*************************************************************************}


procedure TNStack.PushItem(AItem: Pointer);
begin
  FList.Add(AItem);
end;


{*************************************************************************}
{                         TNQueue methods                                 }
{*************************************************************************}


procedure TNQueue.PushItem(AItem: Pointer);
begin
  FList.Insert(0, AItem);
end;


{*************************************************************************}
{                         TNStringList methods                            }
{*************************************************************************}


function TNStringList.CompareStrings(const S1, S2: string): Long;
begin
  if CaseSensitive then
    Result := CompStrCase(S1, S2)
  else
    Result := CompStr(S1, S2);
end;


{*************************************************************************}
{                         TNSortedStrings methods                         }
{*************************************************************************}


constructor TNSortedStrings.Create;
begin
  inherited Create;
  Sorted := True;
  Duplicates := dupIgnore;
  CaseSensitive := False;
end;


{*************************************************************************}
{                         TNFiler methods                                 }
{*************************************************************************}


constructor TNFiler.Create(AStream: TStream);
begin
  inherited Create;
  FStream := AStream;
  FRootStack := TNStack.Create;
end;

destructor TNFiler.Destroy;
begin
  FRootStack.Free;
  inherited Destroy;
end;

procedure TNFiler.SetRoot(Value: TNPersistent);
begin
  FRootStack.Clear;
  FRoot := Value;
  FLocalRoot := Value;
end;


{*************************************************************************}
{                         TNLog methods                                   }
{*************************************************************************}


constructor TNLog.Create(const AFileName: string);
begin
  inherited Create;
  Name := 'Log File';
  Description := 'Stores information about program flow.';
  FBuffer := TNStringList.Create;
  FBufMax := MaxLogBuffer;
  FLineLength := MaxLogLine;
{$IFDEF LOGTOFILES}
  FWriteToFile := not System.IsLibrary;
{$ELSE}
  FWriteToFile := False;
{$ENDIF}
  FFileName := ExeDir + LogDir + PathDelim +
    ExtractFileName(ChangeFileExt(AFileName, '.' + sLogFileExt));
  if FWriteToFile then
  try
    ForceDirectories(ExtractFileDir(FFileName));
    FStream := TNFileStream.Create(FFileName, fmCreate or fmShareDenyNone);
  except
    // File can not be created
    FWriteToFile := False;
  end;
  FIdent := '';
  FCheckRepeats := True;
  FStartTime := Platform.GetTimeMicroseconds;
  AddStartUp;
end;

destructor TNLog.Destroy;
begin
  AddShutDown;
  Flush;
  FStream.Free;
  inherited Destroy;
end;

procedure TNLog.AddStartup;
begin
end;

procedure TNLog.AddShutdown;
begin
end;

procedure TNLog.IncIdent;
begin
  FIdent := FIdent + sLogIdent;
end;

procedure TNLog.DecIdent;
begin
  Delete(FIdent, Length(FIdent) - Length(sLogIdent) + 1, Length(sLogIdent));
end;

procedure TNLog.Flush;
begin
  if FWriteToFile then
  begin
    FBuffer.SaveToStream(FStream);
    FBuffer.Clear;
  end;
end;

procedure TNLog.Add(const Msg: string);
begin
  if FCheckRepeats and (Msg = FLastMsg) and (FLastMsg[Length(FLastMsg)] <>
    sLogLBChar) then
    Inc(FRepeats)
  else
  begin
    if FRepeats > 0 then
    begin
      FBuffer[FBuffer.Count - 1] := FBuffer[FBuffer.Count - 1] +
        (Format(' (Added %d times)', [FRepeats + 1]));
      FRepeats := 0;
    end;
    FLastMsg := Msg;
    FBuffer.Add(SysUtils.WrapText(FIdent + Msg, sLineBreak + FIdent + sLogIdent,
      [' ', '.', ',', '-', #9], FLineLength));
  end;
  if (FBuffer.Count = FBufMax - 1) and FWriteToFile then Flush;
end;

procedure TNLog.Add(const MsgFmt: string; Args: array of const);
begin
  Add(Format(MsgFmt, Args));
end;

procedure TNLog.AddTime(const Msg: string);
begin
  Add('Time: ' + TimeToStr(GetTime) + ' ' +
    FloatToStrF(Platform.GetTimeMicroseconds - FStartTime, ffNumber, 10, 0) +
    ' ' + Msg);
end;

procedure TNLog.AddError(const Msg: string);
begin
  Add(sError + ': ' + Msg);
end;

procedure TNLog.AddWarning(const Msg: string);
begin
  Add(sWarning + ': ' + Msg);
end;

procedure TNLog.BeginSubList(const Msg: string);
begin
  Add(Msg);
  IncIdent;
  AddLineBreak;
end;

procedure TNLog.EndSubList(const Msg: string);
begin
  AddLineBreak;
  DecIdent;
  Add(Msg);
end;

procedure TNLog.AddLineBreak;
begin
  Add(StringOfChar(sLogLBChar, FLineLength - Length(FIdent)));
end;


{*************************************************************************}
{                         TNKernelLog methods                             }
{*************************************************************************}


procedure TNKernelLog.AddStartup;
begin
  inherited AddStartUp;
  AddLineBreak;
  Add('Engine StartUp');
  Add('Execution started: ' + DateToStr(Date) + ' ' + TimeToStr(GetTime));
  Add('Engine version: ' + VersionToStr(EngineVersion));
  Add('Compiler: ' + CompilerName);
  Add('Compilation mode: ' + {$IFNDEF DEBUG}sRelease{$ELSE}sDebug{$ENDIF});
  Add('Platform: ' + Platform.Name + ' ' + VersionToStr(Platform.Version));
  Add('CPU: %s, %d MHz, %s', [Platform.CPUInfo.Name, Platform.CPUInfo.Frequency,
    CPUFeaturesToStr(Platform.CPUInfo.Features)]);
  Add('Memory (Free/Total): Physical %d/%d kB, PageFile %d/%d kB',
    [Platform.MemoryInfo.PhysicalFree, Platform.MemoryInfo.PhysicalTotal,
    Platform.MemoryInfo.PageFileFree, Platform.MemoryInfo.PageFileTotal]);
  BeginSubList('Messages:')
end;

procedure TNKernelLog.AddShutdown;
  function GetLastedTime: string;
  var
    H, M, S, MS: Long;
    Tot: Int64;
  begin
    Tot := (Platform.GetTimeMicroseconds - FStartTime) div 1000;
    H := Tot div (3600000);
    Tot := Tot - H * 3600000;
    M := Tot div (60000);
    Tot := Tot - M * 60000;
    S := Tot div (1000);
    Tot := Tot - S * 1000;
    MS := Tot;
    Result := Format('%.2d:%.2d:%.2d:%.3d', [H, M, S, MS]);
  end;

begin
  EndSubList('Engine ShutDown');
  Add('Class instances created: %d', [GetCreatedObjects]);
  Add('Class instances still living: %d', [GetLivingObjects]);
  Add('Class instances freed: %d', [GetFreedObjects]);
  Add('Execution lasted (H:M:S:MS): ' + GetLastedTime);
  Add('Execution stopped: ' + DateToStr(Date) + ' ' + TimeToStr(GetTime));
  AddLineBreak;
  inherited AddShutDown;
end;


{*************************************************************************}
{                          TNFileModule methods                           }
{*************************************************************************}


function TNFileModule.TestFormat(Stream: TNStream): Boolean;
begin
  Result := False;
end;


{*************************************************************************}
{                         TNFileModuleList methods                        }
{*************************************************************************}


constructor TNFileModuleList.Create;
begin
  inherited Create;
  FModules := TNList.Create;
end;

destructor TNFileModuleList.Destroy;
begin
  FModules.Free;
  inherited Destroy;
end;

function TNFileModuleList.GetDefault: TNFileModule;
begin
  if (FDefault = nil) and (FModules.Count > 0) then
    FDefault := TNFileModule(FModules[0]);
  Result := FDefault;
end;

procedure TNFileModuleList.Add(Module: TNFileModule);
begin
  if Module <> nil then
    if FModules.IndexOf(Module) < 0 then
      if FModules.Add(Module) > -1 then
        Log.Add('File format module registered: %s (%s)', [Module.FormatName,
          Module.Extensions]);
end;

procedure TNFileModuleList.Remove(Module: TNFileModule);
var
  I: Long;
begin
  if Module <> nil then
  begin
    I := FModules.IndexOf(Module);
    if I > -1 then
    begin
      FModules.Delete(I);
      if (FDefault = Module) and (FModules.Count > 0) then
        FDefault := TNFileModule(FModules[0]);
    end;
  end;
end;

function TNFileModuleList.Find(Module: TNFileModule): Boolean;
begin
  Result := FModules.IndexOf(Module) > -1;
end;

function TNFileModuleList.Find(const Ext: string): TNFileModule;
var
  I, J: Long;
  List: TNStringList;
  Found: Boolean;

  procedure StrTokenToStrings(const Exts: string);
  var
    Token: string;
    S: string;
  begin
    List.Clear;
    S := Exts;
    while S <> '' do
    begin
      Token := StrToken(S, [',', ';']);
      List.Add(Token);
    end;
  end;

begin
  Result := nil;
  Found := False;
  List := TNStringList.Create;
  for I := 0 to FModules.Count - 1 do
    if PosNoCase(Ext, TNFileModule(FModules[I]).Extensions) > 0 then
    begin
      StrTokenToStrings(TNFileModule(FModules[I]).Extensions);
      for J := 0 to List.Count - 1 do
        if SameStr(Ext, List[J]) then
          Found := True;

      if Found then
      begin
        Result := TNFileModule(FModules[I]);
        Break;
      end;
    end;
  List.Free;
end;


{*************************************************************************}
{                        TNModuleRegister methods                         }
{*************************************************************************}


{$IFNDEF MODLIBRARY}

constructor TNModuleRegister.Create;
begin
  inherited Create;
  FList := TNSortedStrings.Create;
  FLoadCalls := TNList.Create;
  FUnLoadCalls := TNList.Create;
  FDefaultDirectory := ExeDir + ModuleDir + PathDelim;
end;

destructor TNModuleRegister.Destroy;
begin
  UnloadAll;
  FList.Free;
  FLoadCalls.Free;
  FUnLoadCalls.Free;
  inherited Destroy;
end;

procedure TNModuleRegister.UnloadAll;
begin
  while FList.Count > 0 do
    Unload(PModuleEntry(FList.Objects[0]).FileName);
end;

procedure TNModuleRegister.RunLoadCallBacks(Entry: PModuleEntry);
var
  I: Long;
  Done: Boolean;
begin
  Done := False;
  for I := 0 to FLoadCalls.Count - 1 do
  begin
    if Assigned(FLoadCalls[I]) then
      TModLoadCallBack(FLoadCalls[I])(Entry, Done);
    if Done then Break;
  end;
end;

procedure TNModuleRegister.RunUnLoadCallBacks(Entry: PModuleEntry);
var
  I: Long;
  Done: Boolean;
begin
  Done := False;
  for I := 0 to FUnLoadCalls.Count - 1 do
  begin
    if Assigned(FUnLoadCalls[I]) then
      TModUnLoadCallBack(FUnLoadCalls[I])(Entry, Done);
    if Done then Break;
  end;
end;

procedure TNModuleRegister.Load(const FileName: string);
var
  Entry: PModuleEntry;
  FName: string;
  GetID: TGetIDProc;
  GetCount: TGetCountProc;
begin
  Entry := nil;
  if FileExists(FileName) or FileExists(FDefaultDirectory + FileName) then
  begin
    FName := ExtractFileName(FileName);
    if (PosNoCase(sModulePrefix, FName) = 1) then
      if not IsLoaded(FName) then
      try
        New(Entry);
        Entry^.Name := FName;

        if FileExists(FDefaultDirectory + FileName) then
          FName := FDefaultDirectory + FileName
        else
          Fname := FileName;

        Entry^.Handle := LoadLibrary(FName);
        if Entry^.Handle = 0 then
          raise Exception.Create('Error when loading! Library not loaded!');
        Entry^.FileName := Platform.GetModuleFileName(Entry^.Handle);
        @GetID := GetProcAddress(Entry^.Handle, sGetIDProc);
        if not Assigned(GetID) then
          raise Exception.Create('');
        GetID(Entry^.ID);

        @GetCount := GetProcAddress(Entry^.Handle, sGetCountProc);
        if not Assigned(GetCount) then
          raise Exception.Create('');
        GetCount(Entry^.Count);

        FList.AddObject(Entry^.Name, TObject(Entry));
      except
        on E: Exception do
        begin
          if FList.IndexOfObject(TObject(Entry)) > -1 then
            FList.Delete(FList.IndexOfObject(TObject(Entry)));
          Dispose(Entry);
          // do not show errors here
          Exit;
        end;
      end;
    Log.Add('Module library loaded: ' + Entry^.Name);
    RunLoadCallBacks(Entry);
  end;
end;

procedure TNModuleRegister.LoadFromDirectory(const Dir: string);
var
  DName: string;
  Rec: TSearchRec;
  Res: Long;
begin
  DName := Dir + PathDelim + sModulePrefix + '*';
  ZeroMemory(@Rec, SizeOf(Rec));
  Res := FindFirst(DName, faAnyFile, Rec);
  try
    while Res = 0 do
    begin
      if IsValidModuleLibrary(Dir + PathDelim + Rec.Name) then
        Load(Dir + PathDelim + Rec.Name);
      Res := FindNext(Rec);
    end;
  finally
    FindClose(Rec);
  end;
end;

procedure TNModuleRegister.Unload(const FileName: string);
var
  Entry: PModuleEntry;
  Idx: Long;
begin
  if FList.Find(ExtractFileName(FileName), Idx) then
  begin
    Entry := PModuleEntry(FList.Objects[Idx]);
    RunUnLoadCallBacks(Entry);
    Log.Add('Module library unloaded: ' + Entry^.Name);
    try
      FList.Delete(Idx);
      UnloadLibrary(Entry^.Handle);
      Dispose(Entry);
    except
      on E: Exception do
      begin
        ShowError(ErrorStr(Self, 'Load', 'Error when unloading module library '
          + ' file "' + FileName + '".' + sLineBreak + sExceptMsg +
          E.Message), E.ClassName);
        Exit;
      end;
    end;
  end;
end;

procedure TNModuleRegister.AddLoadCallBack(Call: TModLoadCallBack);
begin
  FLoadCalls.Add(@Call);
end;

procedure TNModuleRegister.AddUnLoadCallBack(Call: TModUnLoadCallBack);
begin
  FUnLoadCalls.Add(@Call);
end;

function TNModuleRegister.IsLoaded(const FileName: string): Boolean;
begin
  Result := FList.IndexOf(ExtractFileName(FileName)) > -1;
end;

function TNModuleRegister.IsValidModuleLibrary(const FileName: string): Boolean;
begin
  Result := Platform.IsExecutable(FileName);
end;
{$ENDIF}


{*************************************************************************}
{                         Internal stuff                                  }
{*************************************************************************}


procedure CheckUnLoadClass(Entry: PModuleEntry; var Done: Boolean);
var
  I: Long;
begin
  I := 0;
  while I < RegClasses.Count - 1 do
  begin
    if FindHInstance(RegClasses.Objects[I]) = Entry^.Handle then
    begin
      UnRegisterClass(TNPersistentClass(RegClasses.Objects[I]));
      Dec(I);
    end;
    Inc(I);
  end;
end;


initialization

  RegisterClass(TNPersistent);
{$IFNDEF MODLIBRARY}
  {This procedure checks if some of TNpersistent classes registered
   in RegClasses are from loaded library (which is now unloading).
   If it is from library then it is unregistered automatically.}
  ModuleRegister.AddUnLoadCallBack(CheckUnLoadClass);
{$ENDIF}

finalization

{$IFNDEF MODLIBRARY}
  FreeAndNil(ModRegInstance);
{$ENDIF}
  FreeAndNil(RegClasses);
  FreeAndNil(LogInstance);

end.

