{*************************************************************************}
{*   PlatformLinux                                                       *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\Linux\PlatformLinux.pas                     *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12th August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains descendant of THPlatform designed for Linux platform.}

unit PlatformLinux;

{$I 'Compiler.inc'}

interface

uses
  Kernel, Objects, Platforms, LibC, LibcElf, LinuxSysInfo, Xlib, CrtLinux;

type


{*************************************************************************}
{                          Class Types                                    }
{*************************************************************************}


  {As you see Linux platform is not extensive now. It will
  remain like this until I buy new HDD and make some space for
  Linux instalation.}
  TNPlatformLinux = class(TNPlatform)
  private
    FTimeVal: TTimeVal;
  protected
    procedure GatherStaticInfo; override;
    procedure GatherDynamicInfo(Index: Long); override;
  public
    procedure MsgBox(const Msg: string); override;
    procedure MsgBox(Number: Int64); override;
    procedure MsgBox(const FormatStr: string; Args: array of const); override;
    function ErrorMsg(const Msg: string): TErrorMsgResult; override;

    function GetModuleFileName(Handle: THandle): string; override;
    function LoadLibrary(const FileName: string): THandle; override;
    function UnloadLibrary(Handle: THandle): Boolean; override;
    function GetProcAddress(Handle: THandle; const ProcName: string): Pointer;
      override;
    function IsExecutable(const FileName: string): Boolean; override;

    function GetTimeMicroseconds: Int64; override;
  end;


{*************************************************************************}
{                         IMPLEMENTATION                                  }
{*************************************************************************}


implementation

uses
  SysUtils;


{*************************************************************************}
{                          TNPlatformLinux methods                        }
{*************************************************************************}


procedure TNPlatformLinux.GatherStaticInfo;
var
  S: string;
begin
  LinuxSysInfo.GetCPUInfo(FCPUInfo);
  LinuxSysInfo.GetOSInfo(FVersion, S);
  Name := S;
end;

procedure TNPlatformLinux.GatherDynamicInfo(Index: Long);
begin
  case Index of
    1: LinuxSysInfo.GetMemoryInfo(FMemoryInfo);
  end;
end;

procedure TNPlatformLinux.MsgBox(const Msg: string);
begin
{$IFDEF CONSOLE}
  WriteLn(#32);
  CrtLinux.TextColor(LightCyan);
  WriteLn('Application Message');
  CrtLinux.TextColor(LightGray);
  WriteLn(Msg + sLineBreak + 'Press [Enter] to continue');
  repeat
  until CrtLinux.ReadKey = #13;
{$ENDIF}
end;

procedure TNPlatformLinux.MsgBox(Number: Int64);
begin
  MsgBox(IntToStr(Number));
end;

procedure TNPlatformLinux.MsgBox(const FormatStr: string;
  Args: array of const);
begin
  MsgBox(SysUtils.Format(FormatStr, Args));
end;

function TNPlatformLinux.ErrorMsg(const Msg: string): TErrorMsgResult;
var
  S: string;
begin
  Result := emContinue;
{$IFDEF CONSOLE}
  WriteLn(#32);
  CrtLinux.TextColor(LightRed);
  WriteLn('Application Error!');
  CrtLinux.TextColor(LightGray);
  WriteLn(Msg + sLineBreak +
    'Press [Enter] to continue or [Esc] to stop execution.');
  S := CrtLinux.ReadKey;
  if S[1] = #27 then
    Result := emTerminate;
{$ENDIF}
end;

function TNPlatformLinux.GetModuleFileName(Handle: THandle): string;
var
  FileName: array[0..FILENAME_MAX] of Char;
begin
  SetString(Result, FileName,
    System.GetModuleFileName(Handle, FileName, SizeOf(FileName)));
end;

function TNPlatformLinux.LoadLibrary(const FileName: string): THandle;
begin
  Result := THandle(dlopen(PChar(FileName), RTLD_LAZY));
end;

function TNPlatformLinux.UnloadLibrary(Handle: THandle): Boolean;
begin
  Result := LongBool(dlclose(Pointer(Handle)));
  InvalidateModuleCache;
end;

function TNPlatformLinux.GetProcAddress(Handle: THandle;
  const ProcName: string): Pointer;
begin
  Result := SysUtils.GetProcAddress(Handle, PChar(ProcName));
end;

function TNPlatformLinux.IsExecutable(const FileName: string): Boolean;
var
  Stream: TNFileStream;
  Header: Elf32_EHdr;
begin
  Result := False;
  if not FileExists(FileName) then Exit;
  try
    Stream := TNFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      Stream.Read(Header, SizeOf(Header));
      if (PULong(@Header.e_ident)^ = ELFMAG) and (Header.e_version in [0, 1, 2])
        then
        Result := True;
    finally
      Stream.Free;
    end;
  except
    Result := False;
  end;
end;

function TNPlatformLinux.GetTimeMicroseconds: Int64;
begin
  GetTimeOfDay(FTimeVal, nil);
  Result := FTimeVal.tv_sec * 1000000 + FTimeVal.tv_usec;
end;

end.

