{*************************************************************************}
{*   PlatformWin32                                                       *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Platforms\Win32\PlatformWin32.pas                     *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 12th August 2003  (12.08.2003)                             }
{*************************************************************************}

{This unit contains descendant of TNPlatform designed for Win32 platform.}

unit PlatformWin32;

{$I 'Compiler.inc'}

interface

uses
  SysUtils, Kernel, Objects, Platforms, Windows, MMSystem,
  Crt, Win32SysInfo;


{*************************************************************************}
{                          Class Types                                    }
{*************************************************************************}


type

  {Descendant of TNPlatform designed for Win32 systems. It should
  support all windows versions from Win95. However, Win2000 or WinXP
  are strongly recommended. Details on class members seek in the declaration
  of TNPlatform class.}
  TNPlatformWin32 = class(TNPlatform)
  private
    FPerfFrequency: Int64;
    FInvPerfFrequency: Float;
    FTimeInt64: Int64;
  protected
    procedure GatherStaticInfo; override;
    procedure GatherDynamicInfo(Index: Long); override;
  public
    procedure MsgBox(const Msg: string); override;
    procedure MsgBox(Number: Int64); override;
    procedure MsgBox(const FormatStr: string; Args: array of const); override;
    function ErrorMsg(const Msg: string): TErrorMsgResult; override;

    function GetModuleFileName(Handle: THandle): string; override;
    function LoadLibrary(const FileName: string): THandle; override;
    function UnloadLibrary(Handle: THandle): Boolean; override;
    function GetProcAddress(Handle: THandle; const ProcName: string): Pointer;
      override;
    function IsExecutable(const FileName: string): Boolean; override;

    function GetTimeMicroseconds: Int64; override;
  end;


{*************************************************************************}
{                         IMPLEMENTATION                                  }
{*************************************************************************}


implementation


{*************************************************************************}
{                         TNPlatformWin32 methods                         }
{*************************************************************************}


procedure TNPlatformWin32.GatherStaticInfo;
var
  S: string;
begin
  Win32SysInfo.GetCPUInfo(FCPUInfo);
  Win32SysInfo.GetOSInfo(FVersion, S);
  Name := S;
  Windows.QueryPerformanceFrequency(FPerfFrequency);
  FInvPerfFrequency := 1 / FPerfFrequency;
end;

procedure TNPlatformWin32.GatherDynamicInfo(Index: Long);
begin
  case Index of
    1: Win32SysInfo.GetMemoryInfo(FMemoryInfo);
  end;
end;


procedure TNPlatformWin32.MsgBox(const Msg: string);
begin
  if not System.IsConsole then
    MessageBox(GetForegroundWindow, PChar(Msg), 'Message',
      MB_ICONINFORMATION or MB_OK)
  else
  begin
    WriteLn;
    Crt.TextColor(LightCyan);
    WriteLn('Application Message');
    Crt.TextColor(LightGray);
    Write(Msg + sLineBreak + 'Press [Enter] to continue');
    repeat
    until Crt.ReadKey = #13;
  end;
end;

procedure TNPlatformWin32.MsgBox(Number: Int64);
begin
  MsgBox(IntToStr(Number));
end;

procedure TNPlatformWin32.MsgBox(const FormatStr: string; Args: array of const);
begin
  MsgBox(SysUtils.Format(FormatStr, Args));
end;

function TNPlatformWin32.ErrorMsg(const Msg: string): TErrorMsgResult;
var
  MRes, Flags: ULong;
  S: string;
begin
  Result := emContinue;
  if IsConsole then
  begin
    Crt.TextColor(LightRed);
    WriteLn;
    WriteLn(sLineBreak + 'Application Error!');
    Crt.TextColor(LightGray);
    Write(Msg + sLineBreak +
      'Press [Enter] to continue or [Esc] to stop execution.');
    S := Crt.ReadKey;
    if S[1] = #27 then
      Result := emTerminate;
  end
  else
  begin
    Flags := MB_ICONERROR or MB_YESNO;
    S := sLineBreak + 'Continue execution?' + sLineBreak;
    MRes := Windows.MessageBox(GetForegroundWindow, PChar(Msg + S),
      'Application Error', Flags);
    if MRes = IDNO then
      Result := emTerminate;
  end;
end;

function TNPlatformWin32.GetModuleFileName(Handle: THandle): string;
var
  FileName: array[0..MAX_PATH] of Char;
begin
  SetString(Result, FileName,
    Windows.GetModuleFileName(Handle, FileName, SizeOf(FileName)));
end;

function TNPlatformWin32.LoadLibrary(const FileName: string): THandle;
begin
  Result := Windows.LoadLibrary(PChar(FileName));
end;

function TNPlatformWin32.UnloadLibrary(Handle: THandle): Boolean;
begin
  Result := Windows.FreeLibrary(Handle);
end;

function TNPlatformWin32.GetProcAddress(Handle: THandle;
  const ProcName: string): Pointer;
begin
  Result := Windows.GetProcAddress(Handle, PChar(ProcName));
end;

function TNPlatformWin32.IsExecutable(const FileName: string): Boolean;
var
  Stream: TNFileStream;
  Image: TImageDosHeader;
  NtHeaders: TImageNtHeaders;
begin
  Result := False;
  if not FileExists(FileName) then Exit;
  try
    Stream := TNFileStream.Create(FileName, fmOpenRead or fmShareDenyNone);
    try
      Stream.Read(Image, SizeOf(Image));
      if Image.e_magic = IMAGE_DOS_SIGNATURE then
      begin
        Stream.Position := Image._lfanew;
        Stream.Read(NtHeaders, SizeOf(NtHeaders));
        if NtHeaders.Signature = IMAGE_NT_SIGNATURE then Result := True;
      end;
    finally
      Stream.Free;
    end;
  except
    Result := False;
  end;
end;

function TNPlatformWin32.GetTimeMicroseconds: Int64;
begin
  QueryPerformanceCounter(FTimeInt64);
  Result := Round(1000000 * FInvPerfFrequency * FTimeInt64);
end;

end.

