{*************************************************************************}
{*   Streaming                                                           *}
{*   Copyright (C) 2003 Marek Mauder, Vampyre SoftWorks                  *}
{*   Engine\Kernel\Streaming.pas                                         *}
{*************************************************************************}
{                                                                         }
{ This program is free software; you can redistribute it and/or modify it }
{ under the terms of the GNU General Public License as published by the   }
{ Free Software Foundation; either version 2 of the License, or (at your  }
{ option) any later version. This program is distributed in the hope that }
{ it will be useful, but WITHOUT ANY WARRANTY; without even the implied   }
{ warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See   }
{ the GNU General Public License for more details.                        }
{                                                                         }
{ You should have received a copy of the GNU General Public License       }
{ along with this program; if not, write to the Free Software Foundation, }
{ Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.         }
{                                                                         }
{*************************************************************************}
{  Created On: 6th September 2003  (06.09.2003)                           }
{*************************************************************************}

{This unit contains abstract classes for streaming of TNPersistent objects
 as well as their binary and XML descendants.}

unit Streaming;

{DONE -oGalfar : Read methods}
{DONE -oGalfar : Add comments and cosmetics}
{DONE -oGalfar : catch exceptions - testing damaged files}
{TODO -oGalfar : support for storing children}

{$I 'Compiler.inc'}

interface

uses
  SysUtils, Classes, Kernel, Objects, Platforms, TypInfo, DOM;

const

  sListProperties = 'Properties';
  sListChildren = 'Children';

type

  {Class for reading TNPersistent objects or other data from
   associated stream.}
  TNReader = class(TNAbstractReader)
  protected
    FPropName: string;

    function BeginWriting: Boolean; virtual; abstract;
    function EndWriting: Boolean; virtual; abstract;
    function BeginObject(var AClassName: string): Boolean; virtual; abstract;
    function EndObject: Boolean; virtual; abstract;
    function BeginProperty(var Name: string): Boolean; virtual; abstract;
    function EndProperty: Boolean; virtual; abstract;
    function BeginCustom: Boolean; virtual; abstract;
    function EndCustom: Boolean; virtual; abstract;

    function IsBeginObject: Boolean; virtual; abstract;
    function IsEndObject: Boolean; virtual; abstract;
    function IsEndList: Boolean; virtual; abstract;
    function IsEndCustom: Boolean; virtual; abstract;

    procedure ReadProperties(Instance: TNPersistent);
    procedure ReadProperty(Instance: TNPersistent);
    procedure ReadPropValue(Instance: TNPersistent; PropInfo: PPropInfo);
    function ReadSet(PropInfo: PPropInfo): Long; virtual; abstract;
    function ReadMethod: string; virtual; abstract;
  public
    procedure DefineProperty(const Name: string; ReadData: TReadDataProc;
      WriteData: TWriteDataProc); override;
    {Reads root object with header and all subobjects and
     properties. It returns nil if fails otherwise returns newly created
     instance of class read from stream.}
    procedure ReadObject(var Instance: TNPersistent);
    procedure ReadInstance(var Instance: TNPersistent); override;

    {Skips next value.}
    procedure SkipValue; virtual; abstract;
  end;

  {Writes TNPersistent objects or other data into associated stream.}
  TNWriter = class(TNAbstractWriter)
  protected
    procedure BeginWriting; virtual; abstract;
    procedure EndWriting; virtual; abstract;
    procedure BeginObject(Instance: TNPersistent); virtual; abstract;
    procedure EndObject; virtual; abstract;
    procedure BeginProperty(const Name: string); virtual; abstract;
    procedure EndProperty; virtual; abstract;
    procedure BeginCustom; virtual; abstract;
    procedure EndCustom; virtual; abstract;

    procedure WriteProperties(Instance: TNPersistent);
    procedure WriteProperty(Instance: TNPersistent; PropInfo: PPropInfo);
    procedure WriteSet(Value: Long; PropInfo: PPropInfo); virtual; abstract;
    procedure WriteMethod(const Value: string); virtual; abstract;
  public
    procedure DefineProperty(const Name: string; ReadData: TReadDataProc;
      WriteData: TWriteDataProc); override;
    {Writes root object with header and all subobjects and
     properties.}
    procedure WriteObject(Instance: TNPersistent);
    procedure WriteInstance(Instance: TNPersistent); override;
  end;

type

  {Type used by binary reader and writer to define value type.}
  TTagType = Byte;

const

  {Constants for TTagType.}
  ttNull = 0;
  ttShort = 1;
  ttByte = 2;
  ttSmall = 3;
  ttWord = 4;
  ttLong = 5;
  ttULong = 6;
  ttInt64 = 7;
  ttString = 8;
  ttLString = 9;
  ttWString = 10;
  ttBinary = 11;
  ttSingle = 12;
  ttDouble = 13;
  ttNil = 14;
  ttBoolean = 15;
  ttListBegin = 16;
  ttListEnd = 17;
  ttObjectBegin = 18;
  ttObjectEnd = 19;
  ttIdent = 20;
  ttSet = 21;
  ttChar = 22;
  ttTrue = 23;
  ttFalse = 24;
  ttMethod = 25;
  ttProperty = 26;
  ttCustomBegin = 27;
  ttCustomEnd = 28;
  ttInvalid = 255;

  {Object header for binary reader and writer.}
  sBinHeader = 'NBinaryObject';
  {Default buffer size for binary reader and writer.}
  BinBufferSize = 8192;

type

  {Descendant of TNReader which reads objects and data stored
   in binary form. Its function is similar to Classes.TReader class.}
  TNBinaryReader = class(TNReader)
  protected
    FBuffer: Pointer;
    FBufSize: Long;
    FBufPos: Long;
    FBufEnd: Long;

    function BeginWriting: Boolean; override;
    function EndWriting: Boolean; override;
    function BeginObject(var AClassName: string): Boolean; override;
    function EndObject: Boolean; override;
    function BeginProperty(var Name: string): Boolean; override;
    function EndProperty: Boolean; override;
    function BeginCustom: Boolean; override;
    function EndCustom: Boolean; override;

    function IsBeginObject: Boolean; override;
    function IsEndObject: Boolean; override;
    function IsEndList: Boolean; override;
    function IsEndCustom: Boolean; override;

    function GetPosition: Long;
    procedure SetPosition(Value: Long);

    function ReadSet(PropInfo: PPropInfo): Long; override;
    function ReadMethod: string; override;

    function ReadTag: TTagType;
    function ReadShortStr: string;
    function NextTag: TTagType;
    procedure SkipBytes(Count: Long);
    procedure ReadBuffer;
    procedure Read(var Buf; Count: Long);
  public
    constructor Create(AStream: TNStream); override;
    destructor Destroy; override;

    procedure Flush; override;

    function BeginList(const Name: string): Boolean; override;
    function EndList: Boolean; override;

    function ReadString: string; override;
    function ReadIdent: string; override;
    function ReadChar: Char; override;
    function ReadLong: Long; override;
    function ReadULong: ULong; override;
    function ReadInt64: Int64; override;
    function ReadBool: Boolean; override;
    function ReadSingle: Single; override;
    function ReadDouble: Double; override;
    procedure ReadBinary(var Data: Pointer; var Count: Long); override;

    procedure SkipValue; override;

    property Position: Long read GetPosition write SetPosition;
  end;

  {Descendant of TNWriter which writes objects and data in binary form.
   Its function is similar to Classes.TWriter class.}
  TNBinaryWriter = class(TNWriter)
  protected
    FBuffer: Pointer;
    FBufSize: Long;
    FBufPos: Long;
    FBufEnd: Long;

    procedure BeginWriting; override;
    procedure EndWriting; override;
    procedure BeginObject(Instance: TNPersistent); override;
    procedure EndObject; override;
    procedure BeginProperty(const Name: string); override;
    procedure EndProperty; override;
    procedure BeginCustom; override;
    procedure EndCustom; override;

    procedure WriteSet(Value: Long; PropInfo: PPropInfo); override;
    procedure WriteMethod(const Value: string); override;

    procedure WriteTag(Tag: TTagType);
    procedure WriteShortStr(const Value: string);
    procedure WriteBuffer;
    procedure Write(const Buf; Count: Long);
  public
    constructor Create(AStream: TNStream); override;
    destructor Destroy; override;

    procedure Flush; override;

    procedure BeginList(const Name: string); override;
    procedure EndList; override;

    procedure WriteString(const Value: string); override;
    procedure WriteIdent(const Value: string); override;
    procedure WriteChar(const Value: Char); override;
    procedure WriteInt(const Value: Long); override;
    procedure WriteInt(const Value: ULong); override;
    procedure WriteInt(const Value: Int64); override;
    procedure WriteBool(const Value: Boolean); override;
    procedure WriteFloat(const Value: Single); override;
    procedure WriteFloat(const Value: Double); override;
    procedure WriteBinary(Data: Pointer; Count: Long); override;
  end;

const

  {Main tag name used by XML reader and writer.}
  sXMLHeader = 'NXMLPersistentObject';

  {Constants defining value type.}
  sXString = 'String';
  sXLong = 'Long';
  sXULong = 'ULong';
  sXInt64 = 'Int64';
  sXSingle = 'Single';
  sXDouble = 'Double';
  sXChar = 'Char';
  sXSet = 'Set';
  sXIdent = 'Ident';
  sXBool = 'Bool';
  sXMethod = 'Method';
  sXBinary = 'Binary';
  sXCustom = 'Custom';

  {Constants defining node attribute name for XML reader and writer.}
  sXName = 'Name';
  sXType = 'Type';
  sXValue = 'Value';
  sXVariable = 'Var';
  sXProperty = 'Property';
  sXClass = 'Class';
  sXObject = 'Object';
  sXList = 'List';
  sXSize = 'Size';

type

  {Descendant of TNReader which reads objects and data stored
   in the form of XML document.}
  TNXMLReader = class(TNReader)
  protected
    FDoc: TXMLDocument;
    FLocal: TDOMElement;
    FAttr: TDOMElement;

    function BeginWriting: Boolean; override;
    function EndWriting: Boolean; override;
    function BeginObject(var AClassName: string): Boolean; override;
    function EndObject: Boolean; override;
    function BeginProperty(var Name: string): Boolean; override;
    function EndProperty: Boolean; override;
    function BeginCustom: Boolean; override;
    function EndCustom: Boolean; override;

    function IsBeginObject: Boolean; override;
    function IsEndObject: Boolean; override;
    function IsEndList: Boolean; override;
    function IsEndCustom: Boolean; override;

    function ReadSet(PropInfo: PPropInfo): Long; override;
    function ReadMethod: string; override;

    function CheckLocalType(const TypeStr: string): Boolean;
    function GetLocalStr: string;
    function GetLocalInt: Int64;
    function GetLocalFloat: Double;
    procedure GetLocalBinary(var Data: Pointer; var Count: Long);
  public
    constructor Create(AStream: TNStream); override;
    destructor Destroy; override;

    {Reads entire stream and creates XML tree.}
    procedure Flush; override;

    function BeginList(const Name: string): Boolean; override;
    function EndList: Boolean; override;

    function ReadString: string; override;
    function ReadIdent: string; override;
    function ReadChar: Char; override;
    function ReadLong: Long; override;
    function ReadULong: ULong; override;
    function ReadInt64: Int64; override;
    function ReadBool: Boolean; override;
    function ReadSingle: Single; override;
    function ReadDouble: Double; override;
    procedure ReadBinary(var Data: Pointer; var Count: Long); override;

    procedure SkipValue; override;
  end;

  {Descendant of TNWriter which writes objects and data
   to XML document.}
  TNXMLWriter = class(TNWriter)
  protected
    FDoc: TXMLDocument;
    FLocal: TDOMElement;
    FNode: TDOMElement;
    FPropNode: TDOMElement;
    FStack: TNStack;

    procedure BeginWriting; override;
    procedure EndWriting; override;
    procedure BeginObject(Instance: TNPersistent); override;
    procedure EndObject; override;
    procedure BeginProperty(const Name: string); override;
    procedure EndProperty; override;
    procedure BeginCustom; override;
    procedure EndCustom; override;

    procedure WriteSet(Value: Long; PropInfo: PPropInfo); override;
    procedure WriteMethod(const Value: string); override;

    function GetNewNode(const Name: string = ''; Global: Boolean = False):
      TDOMElement;
  public
    constructor Create(AStream: TNStream); override;
    destructor Destroy; override;

    {Writes entire XML tree to stream.}
    procedure Flush; override;

    procedure BeginList(const Name: string); override;
    procedure EndList; override;

    procedure WriteString(const Value: string); override;
    procedure WriteIdent(const Value: string); override;
    procedure WriteChar(const Value: Char); override;
    procedure WriteInt(const Value: Long); override;
    procedure WriteInt(const Value: ULong); override;
    procedure WriteInt(const Value: Int64); override;
    procedure WriteBool(const Value: Boolean); override;
    procedure WriteFloat(const Value: Single); override;
    procedure WriteFloat(const Value: Double); override;
    procedure WriteBinary(Data: Pointer; Count: Long); override;
  end;


{*************************************************************************}
{                         IMPLEMEMTATION                                  }
{*************************************************************************}


implementation


{*************************************************************************}
{                         TNReader methods                                }
{*************************************************************************}


procedure TNReader.ReadProperties(Instance: TNPersistent);
begin
  while not IsEndList do
    ReadProperty(Instance);
end;

procedure TNReader.ReadProperty(Instance: TNPersistent);
var
  PropInfo: PPropInfo;
begin
  if BeginProperty(FPropName) then
  begin
    PropInfo := GetPropInfo(Instance.ClassInfo, FPropName);
    if PropInfo <> nil then
      ReadPropValue(Instance, PropInfo)
    else
    begin
      Instance.DefineProperties(Self);
      if FPropName <> '' then
        SkipValue;
    end;
    EndProperty;
  end;
end;

procedure TNReader.ReadPropValue(Instance: TNPersistent;
  PropInfo: PPropInfo);

  procedure ReadObjectProp;
  var
    Obj: TNPersistent;
  begin
    Obj := nil;
    if IsBeginObject then
      ReadInstance(Obj)
    else
      ReadIdent;
    SetObjectProp(Instance, PropInfo, Obj);
  end;

  procedure ReadMethodProp;
  var
    S: string;
    Method: TMethod;
    Obj: TNPersistent;
  begin
    S := ReadMethod;
    if S = sNil then Exit;
    Method.Code := Instance.MethodAddress(S);
    Obj := Instance;
    if Method.Code = nil then
      if FLocalRoot <> nil then
      begin
        Method.Code := FLocalRoot.MethodAddress(S);
        Obj := FLocalRoot;
      end;
    if Method.Code = nil then
      if FRoot <> nil then
      begin
        Method.Code := FRoot.MethodAddress(S);
        Obj := FRoot;
      end;
    Method.Data := Obj;
    if Method.Code <> nil then
      SetMethodProp(Instance, PropInfo, Method);
  end;

begin
  if PropInfo^.SetProc = nil then
    SkipValue
  else
    case PropInfo^.PropType^.Kind of
      tkInteger:
        case GetTypeData(PropInfo^.PropType^)^.OrdType of
          otSByte, otSWord, otSLong: SetOrdProp(Instance, PropInfo, ReadLong);
          otUByte, otUWord, otULong: SetOrdProp(Instance, PropInfo, ReadULong);
        end;
      tkInt64:
        SetInt64Prop(Instance, PropInfo, ReadInt64);
      tkFloat:
        case GetTypeData(PropInfo^.PropType^)^.FloatType of
          ftSingle: SetFloatProp(Instance, PropInfo, ReadSingle);
          ftDouble, ftCurr: SetFloatProp(Instance, PropInfo, ReadDouble);
        end;
      tkChar:
        SetOrdProp(Instance, PropInfo, Ord(ReadChar));
      tkEnumeration:
        SetOrdProp(Instance, PropInfo, GetEnumValue(PropInfo^.PropType^,
          ReadIdent));
      tkString, tkLString:
        SetStrProp(Instance, PropInfo, ReadString);
      tkSet:
        SetOrdProp(Instance, PropInfo, ReadSet(PropInfo));
      tkClass:
        ReadObjectProp;
      tkMethod:
        ReadMethodProp;
      else
        SkipValue;
    end;
end;

procedure TNReader.DefineProperty(const Name: string;
  ReadData: TReadDataProc; WriteData: TWriteDataProc);
begin
  if SameStr(Name, FPropName) and Assigned(ReadData) then
  begin
    if BeginCustom then
    begin
      ReadData(Self);
      FPropName := '';
      while not IsEndCustom do
        SkipValue;
      EndCustom;
    end;
  end;
end;

procedure TNReader.ReadObject(var Instance: TNPersistent);
begin
  Instance := nil;
  if BeginWriting then
  begin
    Root := nil;
    ReadInstance(Instance);
    EndWriting;
    Log.Add('%s: Object read (Class: %s)', [Self.ClassName,
      Instance.ClassName]);
  end;
end;

procedure TNReader.ReadInstance(var Instance: TNPersistent);
var
  AClassName: string;
  AClassType: TNPersistentClass;
  OldLocalRoot: TNPersistent;
  OldRoot: TNPersistent;
begin
  Instance := nil;
  OldRoot := nil;
  OldLocalRoot := nil;
  if BeginObject(AClassName) then
  begin
    AClassType := GetClass(AClassName);
    if AClassType <> nil then
    try
      Instance := AClassType.Create;
      if BeginList(sListProperties) then
      try
        OldRoot := FRoot;
        OldLocalRoot := FLocalRoot;
        if FRootStack.Count > 1 then
          FLocalRoot := TNPersistent(FRootStack.Peek);
        if FRoot = nil then
          Root := Instance;
        FRootStack.Push(Instance);
        ReadProperties(Instance);
        EndList;
        EndObject;
      finally
        FRootStack.Pop;
        FRoot := OldRoot;
        FLocalRoot := OldLocalRoot;
      end;
    except
      on E: Exception do ShowError(E.Message, E.ClassName);
    end;
  end;
end;


{*************************************************************************}
{                         TNWriter methods                                }
{*************************************************************************}


procedure TNWriter.WriteProperties(Instance: TNPersistent);
var
  I, Count: Long;
  PropInfo: PPropInfo;
  PropList: PPropList;
begin
  Count := GetTypeData(Instance.ClassInfo)^.PropCount;
  if Count > 0 then
  begin
    GetMem(PropList, Count * SizeOf(Pointer));
    try
      GetPropInfos(Instance.ClassInfo, PropList);
      for I := 0 to Count - 1 do
      begin
        PropInfo := PropList^[I];
        if PropInfo = nil then
          Break;
        if IsStoredProp(Instance, PropInfo) then
          WriteProperty(Instance, PropInfo);
      end;
    finally
      FreeMem(PropList, Count * SizeOf(Pointer));
    end;
  end;
  Instance.DefineProperties(Self);
end;

procedure TNWriter.WriteProperty(Instance: TNPersistent;
  PropInfo: PPropInfo);

  procedure WriteOrdProp;
  var
    Value: Long;
  begin
    Value := GetOrdProp(Instance, PropInfo);
    if Value = PPropInfo(PropInfo)^.Default then Exit;
    BeginProperty(PropInfo^.Name);
    case PropInfo^.PropType^.Kind of
      tkInteger:
        case GetTypeData(PropInfo^.PropType^)^.OrdType of
          otUByte, otUWord, otULong:
            WriteInt(ULong(Value));
          otSByte, otSWord, otSLong:
            WriteInt(Long(Value));
        end;
      tkChar:
        WriteChar(Chr(Value));
      tkSet:
        WriteSet(Value, PropInfo);
      tkEnumeration:
        WriteIdent(GetEnumName(PropInfo^.PropType^, Value));
    end;
    EndProperty;
  end;

  procedure WriteInt64Prop;
  var
    Value: Int64;
  begin
    Value := GetInt64Prop(Instance, PropInfo);
    if Value = PPropInfo(PropInfo)^.Default then Exit;
    BeginProperty(PropInfo^.Name);
    WriteInt(Value);
    EndProperty;
  end;

  procedure WriteFloatProp;
  var
    ValS: Single;
    ValD: Double;
  begin
    BeginProperty(PropInfo^.Name);
    ValD := GetFloatProp(Instance, PropInfo);
    if (GetTypeData(PropInfo^.PropType^)^.FloatType = ftSingle) then
    begin
      ValS := ValD;
      WriteFloat(ValS);
    end
    else
      WriteFloat(ValD);
    EndProperty;
  end;

  procedure WriteMethodProp;
  var
    Value: TMethod;
    Name: string;
  begin
    Value := GetMethodProp(Instance, PropInfo);
    BeginProperty(PropInfo^.Name);
    if Value.Code = nil then
      WriteIdent(sNil)
    else
    begin
      if FRoot <> nil then
        Name := FRoot.MethodName(Value.Code)
      else
        Name := '';
      if Name = '' then
        if FLocalRoot <> nil then
          Name := FLocalRoot.MethodName(Value.Code)
        else
          Name := '';
      if Name = '' then
        Name := Instance.MethodName(Value.Code);
      if Name = '' then
        WriteIdent(sNil)
      else
        WriteMethod(Name);
    end;
    EndProperty;
  end;

  procedure WriteIntfProp;
  begin

  end;

  procedure WriteObjectProp;
  var
    Obj: TNPersistent;
  begin
    BeginProperty(PropInfo^.Name);
    Obj := TNPersistent(TypInfo.GetObjectProp(Instance, PropInfo));
    if Obj <> nil then
      WriteInstance(Obj)
    else
      WriteIdent(sNil);
    EndProperty;
  end;

begin
  if (PropInfo^.GetProc <> nil) and (PropInfo^.SetProc <> nil) then
    case PropInfo^.PropType^.Kind of
      tkInteger, tkChar, tkEnumeration, tkSet:
        WriteOrdProp;
      tkInt64:
        WriteInt64Prop;
      tkString, tkLString:
        begin
          BeginProperty(PropInfo^.Name);
          WriteString(GetStrProp(Instance, PropInfo));
          EndProperty;
        end;
      tkFloat:
        WriteFloatProp;
      tkMethod:
        WriteMethodProp;
      tkInterface:
        WriteIntfProp;
      tkClass:
        WriteObjectProp;
    end;
end;

procedure TNWriter.DefineProperty(const Name: string;
  ReadData: TReadDataProc; WriteData: TWriteDataProc);
begin
  if Assigned(WriteData) then
  begin
    BeginProperty(Name);
    BeginCustom;
    WriteData(Self);
    EndCustom;
    EndProperty;
  end;
end;

procedure TNWriter.WriteObject(Instance: TNPersistent);
begin
  if Instance <> nil then
  begin
    Root := Instance;
    BeginWriting;
    WriteInstance(Instance);
    EndWriting;
    Log.Add('%s: Object written (Class: %s)', [Self.ClassName,
      Instance.ClassName]);
    Root := nil;
  end;
end;

procedure TNWriter.WriteInstance(Instance: TNPersistent);
var
  OldLocalRoot: TNPersistent;
  OldRoot: TNPersistent;
begin
  if Instance = nil then Exit;
  if Instance.ClassInfo = nil then Exit;
  OldRoot := nil;
  OldLocalRoot := nil;
  try
    try
      OldLocalRoot := FLocalRoot;
      OldRoot := FRoot;
      if FRootStack.Count > 1 then
        FLocalRoot := TNPersistent(FRootStack.Peek);
      if FRoot = nil then
        Root := Instance;
      FRootStack.Push(Instance);
      BeginObject(Instance);
      BeginList(sListProperties);
      WriteProperties(Instance);
      EndList;
      EndObject;
    finally
      FRootStack.Pop;
      FRoot := OldRoot;
      FLocalRoot := OldLocalRoot;
    end;
  except
    on E: Exception do ShowError(E.Message, E.ClassName);
  end;
end;


{*************************************************************************}
{                      TNBinaryReader methods                             }
{*************************************************************************}


constructor TNBinaryReader.Create(AStream: TNStream);
begin
  inherited Create(AStream);
  FBufSize := BinBufferSize;
  GetMem(FBuffer, FBufSize);
end;

destructor TNBinaryReader.Destroy;
begin
  if FBuffer <> nil then
    FreeMem(FBuffer);
  inherited Destroy;
end;

function TNBinaryReader.BeginWriting: Boolean;
begin
  Result := Iff(ReadShortStr = sBinHeader, True, False);
end;

function TNBinaryReader.EndWriting: Boolean;
begin
  Result := True;
end;

function TNBinaryReader.BeginObject(var AClassName: string): Boolean;
begin
  AClassName := '';
  if NextTag = ttObjectBegin then
  begin
    ReadTag;
    Result := True;
    AClassName := ReadShortStr;
  end
  else
    Result := False;
end;

function TNBinaryReader.EndObject: Boolean;
begin
  Result := ReadTag = ttObjectEnd;
end;

function TNBinaryReader.BeginProperty(var Name: string): Boolean;
begin
  Name := '';
  if NextTag = ttProperty then
  begin
    Result := True;
    ReadTag;
    Name := ReadShortStr;
  end
  else
    Result := False;
end;

function TNBinaryReader.EndProperty: Boolean;
begin
  Result := True;
end;

function TNBinaryReader.BeginCustom: Boolean;
begin
  if NextTag = ttCustomBegin then
  begin
    Result := True;
    ReadTag;
  end
  else
    Result := False;
end;

function TNBinaryReader.EndCustom: Boolean;
begin
  Result := ReadTag = ttCustomEnd;
end;

function TNBinaryReader.IsBeginObject: Boolean;
begin
  Result := NextTag = ttObjectBegin;
end;

function TNBinaryReader.IsEndObject: Boolean;
begin
  Result := NextTag = ttObjectEnd;
end;

function TNBinaryReader.IsEndList: Boolean;
begin
  Result := NextTag = ttListEnd;
end;

function TNBinaryReader.IsEndCustom: Boolean;
begin
  Result := NextTag = ttCustomEnd;
end;

function TNBinaryReader.GetPosition: Long;
begin
  Result := FStream.Position - (FBufEnd - FBufPos);
end;

procedure TNBinaryReader.SetPosition(Value: Long);
begin
  FStream.Position := Value;
  FBufPos := 0;
  FBufEnd := 0;
end;

function TNBinaryReader.ReadSet(PropInfo: PPropInfo): Long;
var
  EnumType: PTypeInfo;
  EnumName: string;
begin
  if NextTag = ttSet then
  begin
    ReadTag;
    EnumType := GetTypeData(PropInfo^.PropType^)^.CompType^;
    Result := 0;
    while True do
    begin
      EnumName := ReadShortStr;
      if EnumName = '' then Break;
      Include(TIntegerSet(Result), GetEnumValue(EnumType, EnumName));
    end;
  end
  else
    Result := 0;
end;

function TNBinaryReader.ReadTag: TTagType;
begin
  Read(Result, SizeOf(TTagType));
end;

function TNBinaryReader.ReadShortStr: string;
var
  L: Long;
begin
  Read(L, SizeOf(Byte));
  SetLength(Result, L);
  Read(Pointer(Result)^, L);
end;

function TNBinaryReader.NextTag: TTagType;
begin
  Result := ReadTag;
  Dec(FBufPos);
end;

procedure TNBinaryReader.SkipBytes(Count: Long);
var
  Bytes: array[0..255] of Char;
begin
  while Count > 0 do
    if Count > SizeOf(Bytes) then
    begin
      Read(Bytes, SizeOf(Bytes));
      Dec(Count, SizeOf(Bytes));
    end
    else
    begin
      Read(Bytes, Count);
      Count := 0;
    end;
end;

procedure TNBinaryReader.ReadBuffer;
begin
  try
    FBufEnd := FStream.Read(FBuffer^, FBufSize);
    FBufPos := 0;
  except
    raise Exception.Create(ErrorStr(Self, 'ReadBuffer',
      'Error when reading from stream.'));
  end;
  if FBufEnd = 0 then
    raise Exception.Create(ErrorStr(Self, 'ReadBuffer',
    'Error when reading from stream.'));
end;

procedure TNBinaryReader.Read(var Buf; Count: Long);
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        MOV     EDI,EDX
        MOV     EBX,ECX
        MOV     ESI,EAX
        JMP     @@6
@@1:    MOV     ECX,[ESI].TNBinaryReader.FBufEnd
        SUB     ECX,[ESI].TNBinaryReader.FBufPos
        JA      @@2
        MOV     EAX,ESI
        CALL    TNBinaryReader.ReadBuffer
        MOV     ECX,[ESI].TNBinaryReader.FBufEnd
@@2:    CMP     ECX,EBX
        JB      @@3
        MOV     ECX,EBX
@@3:    PUSH    ESI
        SUB     EBX,ECX
        MOV     EAX,[ESI].TNBinaryReader.FBuffer
        ADD     EAX,[ESI].TNBinaryReader.FBufPos
        ADD     [ESI].TNBinaryReader.FBufPos,ECX
        MOV     ESI,EAX
        MOV     EDX,ECX
        SHR     ECX,2
        CLD
        REP     MOVSD
        MOV     ECX,EDX
        AND     ECX,3
        REP     MOVSB
        POP     ESI
@@6:    OR      EBX,EBX
        JNE     @@1
        POP     EBX
        POP     EDI
        POP     ESI
end;

procedure TNBinaryReader.Flush;
begin
  FStream.Position := Position;
  FBufPos := 0;
  FBufEnd := 0;
end;

function TNBinaryReader.BeginList(const Name: string): Boolean;
begin
  if NextTag = ttListBegin then
  begin
    Result := True;
    ReadTag;
  end
  else
    Result := False;
end;

function TNBinaryReader.EndList: Boolean;
begin
  Result := ReadTag = ttListEnd;
end;

function TNBinaryReader.ReadString: string;
var
  L: Long;
begin
  if NextTag in [ttString, ttLString] then
  begin
    L := 0;
    case ReadTag of
      ttString:
        Read(L, SizeOf(Byte));
      ttLString:
        Read(L, SizeOf(Long));
    end;
    SetLength(Result, L);
    Read(Pointer(Result)^, L);
  end
  else
    Result := '';
end;

function TNBinaryReader.ReadIdent: string;
begin
  if NextTag in [ttIdent, ttFalse, ttTrue, ttNil, ttNull] then
    case ReadTag of
      ttIdent:
        Result := ReadShortStr;
      ttFalse:
        Result := sFalse;
      ttTrue:
        Result := sTrue;
      ttNil:
        Result := sNil;
      ttNull:
        Result := sNull;
    end
  else
    Result := '';
end;

function TNBinaryReader.ReadChar: Char;
begin
  if NextTag = ttChar then
  begin
    ReadTag;
    Read(Result, SizeOf(Char));
  end
  else
    Result := #0;
end;

function TNBinaryReader.ReadLong: Long;
var
  S: Short;
  I: Small;
begin
  case NextTag of
    ttShort:
      begin
        ReadTag;
        Read(S, SizeOf(Short));
        Result := S;
      end;
    ttSmall:
      begin
        ReadTag;
        Read(I, SizeOf(Small));
        Result := I;
      end;
    ttLong:
      begin
        ReadTag;
        Read(Result, SizeOf(Long));
      end;
    else
      Result := 0;
  end;
end;

function TNBinaryReader.ReadULong: ULong;
var
  B: Byte;
  W: Word;
begin
  case NextTag of
    ttByte:
      begin
        ReadTag;
        Read(B, SizeOf(Byte));
        Result := B;
      end;
    ttWord:
      begin
        ReadTag;
        Read(W, SizeOf(Word));
        Result := W;
      end;
    ttULong:
      begin
        ReadTag;
        Read(Result, SizeOf(ULong));
      end;
    else
      Result := ReadLong;
  end;
end;

function TNBinaryReader.ReadInt64: Int64;
begin
  if NextTag = ttInt64 then
  begin
    ReadTag;
    Read(Result, SizeOf(Int64));
  end
  else
    Result := ReadULong;
end;

function TNBinaryReader.ReadBool: Boolean;
begin
  if NextTag in [ttTrue, ttFalse] then
    Result := ReadTag = ttTrue
  else
    Result := False;
end;

function TNBinaryReader.ReadSingle: Single;
begin
  if NextTag = ttSingle then
  begin
    ReadTag;
    Read(Result, SizeOf(Single));
  end
  else
    Result := 0.0;
end;

function TNBinaryReader.ReadDouble: Double;
begin
  if NextTag = ttDouble then
  begin
    ReadTag;
    Read(Result, SizeOf(Double));
  end
  else
    Result := 0.0;
end;

procedure TNBinaryReader.ReadBinary(var Data: Pointer; var Count: Long);
begin
  if NextTag = ttBinary then
  begin
    ReadTag;
    Read(Count, SizeOf(Long));
    GetMem(Data, Count);
    Read(Data^, Count);
  end
  else
  begin
    Data := nil;
    Count := 0;
  end;
end;

procedure TNBinaryReader.SkipValue;

  procedure SkipBinary(BytesPerUnit: Long);
  var
    Count: Long;
  begin
    Read(Count, SizeOf(Count));
    SkipBytes(Count * BytesPerUnit);
  end;

  procedure SkipSetBody;
  begin
    while ReadShortStr <> '' do
    begin
    end;
  end;

  procedure SkipList;
  begin
    while not IsEndList do
      SkipValue;
    ReadTag;
  end;

  procedure SkipObject;
  begin
    ReadShortStr;
    ReadTag;
    while not IsEndObject do
      SkipList;
    ReadTag;
  end;

  procedure SkipCustom;
  begin
    while not IsEndCustom do
      SkipValue;
    ReadTag;
  end;

  procedure SkipProperty;
  begin
    ReadShortStr;
    SkipValue;
  end;

begin
  case ReadTag of
    ttNull, ttFalse, ttTrue, ttNil, ttListEnd,
      ttObjectEnd, ttCustomEnd: ;
    ttString, ttIdent, ttMethod: ReadShortStr;
    ttLString: SkipBinary(1);
    ttShort, ttByte: SkipBytes(SizeOf(Byte));
    ttSmall, ttWord: SkipBytes(SizeOf(Word));
    ttLong, ttULong: SkipBytes(SizeOf(Long));
    ttInt64: SkipBytes(Sizeof(Int64));
    ttSingle: SkipBytes(SizeOf(Single));
    ttDouble: SkipBytes(SizeOf(Double));
    ttBinary: SkipBinary(1);
    ttSet: SkipSetBody;
    ttListBegin: SkipList;
    ttObjectBegin: SkipObject;
    ttCustomBegin: SkipCustom;
    ttProperty: SkipProperty;
  end;
end;


{*************************************************************************}
{                      TNBinaryWriter methods                             }
{*************************************************************************}


constructor TNBinaryWriter.Create(AStream: TNStream);
begin
  inherited Create(AStream);
  FBufSize := BinBufferSize;
  GetMem(FBuffer, FBufSize);
end;

destructor TNBinaryWriter.Destroy;
begin
  if FBuffer <> nil then
    FreeMem(FBuffer, FBufSize);
  inherited Destroy;
end;

procedure TNBinaryWriter.BeginWriting;
begin
  WriteShortStr(sBinHeader);
end;

procedure TNBinaryWriter.EndWriting;
begin
end;

procedure TNBinaryWriter.BeginObject(Instance: TNPersistent);
begin
  WriteTag(ttObjectBegin);
  WriteShortStr(GetTypeData(PTypeInfo(Instance.ClassType.ClassInfo)).UnitName +
    '.' + Instance.ClassName);
end;

procedure TNBinaryWriter.EndObject;
begin
  WriteTag(ttObjectEnd);
end;

procedure TNBinaryWriter.BeginProperty(const Name: string);
begin
  WriteTag(ttProperty);
  WriteShortStr(Name);
end;

procedure TNBinaryWriter.EndProperty;
begin
end;

procedure TNBinaryWriter.BeginCustom;
begin
  WriteTag(ttCustomBegin);
end;

procedure TNBinaryWriter.EndCustom;
begin
  WriteTag(ttCustomEnd);
end;

procedure TNBinaryWriter.WriteSet(Value: Long; PropInfo: PPropInfo);
var
  I: Long;
  BaseType: PTypeInfo;
begin
  BaseType := GetTypeData(PropInfo^.PropType^)^.CompType^;
  WriteTag(ttSet);
  for I := 0 to SizeOf(TIntegerSet) * 8 - 1 do
    if I in TIntegerSet(Value) then
      WriteShortStr(GetEnumName(BaseType, I));
  WriteShortStr('');
end;

procedure TNBinaryWriter.WriteMethod(const Value: string);
begin
  WriteTag(ttMethod);
  WriteShortStr(Value);
end;

function TNBinaryReader.ReadMethod: string;
begin
  if NextTag = ttMethod then
  begin
    ReadTag;
    Result := ReadShortStr;
  end
  else
    if NextTag = ttIdent then
      Result := ReadIdent;
end;

procedure TNBinaryWriter.WriteTag(Tag: TTagType);
begin
  Write(Tag, SizeOf(TTagType));
end;

procedure TNBinaryWriter.WriteShortStr(const Value: string);
var
  L: Long;
begin
  L := Length(Value);
  if L > 255 then
    L := 255;
  Write(L, SizeOf(Byte));
  Write(Pointer(Value)^, L);
end;

procedure TNBinaryWriter.WriteBuffer;
begin
  try
    FStream.WriteBuffer(FBuffer^, FBufPos);
    FBufPos := 0;
  except
    raise Exception.Create(ErrorStr(Self, 'WriteBuffer',
      'Error when writing stream.'));
  end;
end;

procedure TNBinaryWriter.Write(const Buf; Count: Long);
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        MOV     ESI,EDX
        MOV     EBX,ECX
        MOV     EDI,EAX
        JMP     @@6
@@1:    MOV     ECX,[EDI].TNBinaryWriter.FBufSize
        SUB     ECX,[EDI].TNBinaryWriter.FBufPos
        JA      @@2
        MOV     EAX,EDI
        CALL    TNBinaryWriter.WriteBuffer
        MOV     ECX,[EDI].TNBinaryWriter.FBufSize
@@2:    CMP     ECX,EBX
        JB      @@3
        MOV     ECX,EBX
@@3:    SUB     EBX,ECX
        PUSH    EDI
        MOV     EAX,[EDI].TNBinaryWriter.FBuffer
        ADD     EAX,[EDI].TNBinaryWriter.FBufPos
        ADD     [EDI].TNBinaryWriter.FBufPos,ECX
@@5:    MOV     EDI,EAX
        MOV     EDX,ECX
        SHR     ECX,2
        CLD
        REP     MOVSD
        MOV     ECX,EDX
        AND     ECX,3
        REP     MOVSB
        POP     EDI
@@6:    OR      EBX,EBX
        JNE     @@1
        POP     EBX
        POP     EDI
        POP     ESI
end;

procedure TNBinaryWriter.Flush;
begin
  WriteBuffer;
end;

procedure TNBinaryWriter.BeginList(const Name: string);
begin
  WriteTag(ttListBegin);
end;

procedure TNBinaryWriter.EndList;
begin
  WriteTag(ttListEnd);
end;

procedure TNBinaryWriter.WriteString(const Value: string);
var
  I: Long;
begin
  I := Length(Value);
  if I <= 255 then
  begin
    WriteTag(ttString);
    Write(I, SizeOf(Byte));
  end
  else
  begin
    WriteTag(ttLString);
    Write(I, SizeOf(Long));
  end;
  if I > 0 then
    Write(Pointer(Value)^, I);
end;

procedure TNBinaryWriter.WriteIdent(const Value: string);
begin
  if SameStr(Value, sFalse) then
    WriteTag(ttFalse)
  else
    if SameStr(Value, sTrue) then
      WriteTag(ttTrue)
    else
      if SameStr(Value, sNull) then
        WriteTag(ttNull)
      else
        if SameStr(Value, sNil) then
          WriteTag(ttNil)
        else
        begin
          WriteTag(ttIdent);
          WriteShortStr(Value);
        end;
end;

procedure TNBinaryWriter.WriteChar(const Value: Char);
begin
  WriteTag(ttChar);
  Write(Value, SizeOf(Char));
end;

procedure TNBinaryWriter.WriteInt(const Value: Long);
begin
  if (Value >= MinShort) and (Value <= MaxShort) then
  begin
    WriteTag(ttShort);
    Write(Value, SizeOf(Short));
  end
  else
    if (Value >= MinByte) and (Value <= MaxByte) then
    begin
      WriteTag(ttByte);
      Write(Value, SizeOf(Byte));
    end
    else
      if (Value >= MinSmall) and (Value <= MaxSmall) then
      begin
        WriteTag(ttSmall);
        Write(Value, SizeOf(Small));
      end
      else
      begin
        WriteTag(ttLong);
        Write(Value, SizeOf(Long));
      end;
end;

procedure TNBinaryWriter.WriteInt(const Value: ULong);
begin
  if Value <= MaxByte then
  begin
    WriteTag(ttByte);
    Write(Value, SizeOf(Byte));
  end
  else
    if Value <= MaxWord then
    begin
      WriteTag(ttWord);
      Write(Value, SizeOf(Word));
    end
    else
    begin
      WriteTag(ttULong);
      Write(Value, SizeOf(ULong));
    end;
end;

procedure TNBinaryWriter.WriteInt(const Value: Int64);
begin
  if (Value >= MinLong) and (Value <= MaxLong) then
    WriteInt(Long(Value))
  else
    if (Value >= MinULong) and (Value <= MaxULong) then
      WriteInt(ULong(Value))
    else
    begin
      WriteTag(ttInt64);
      Write(Value, Sizeof(Int64));
    end;
end;

procedure TNBinaryWriter.WriteBool(const Value: Boolean);
begin
  if Value then
    WriteTag(ttTrue)
  else
    WriteTag(ttFalse);
end;

procedure TNBinaryWriter.WriteFloat(const Value: Double);
begin
  WriteTag(ttDouble);
  Write(Value, SizeOf(Double));
end;

procedure TNBinaryWriter.WriteFloat(const Value: Single);
begin
  WriteTag(ttSingle);
  Write(Value, SizeOf(Single));
end;

procedure TNBinaryWriter.WriteBinary(Data: Pointer; Count: Long);
begin
  WriteTag(ttBinary);
  Write(Count, SizeOf(Count));
  Write(Data^, Count);
end;


{*************************************************************************}
{                         TXMLReader methods                              }
{*************************************************************************}


constructor TNXMLReader.Create(AStream: TNStream);
begin
  inherited Create(AStream);
  FDoc := TXMLDocument.Create;
  FDoc.WriteHeaders := False;
end;

destructor TNXMLReader.Destroy;
begin
  FDoc.Free;
  inherited Destroy;
end;

function TNXMLReader.BeginWriting: Boolean;
begin
  FLocal := FDoc.DocumentElement;
  Result := FLocal <> nil;
  if Result then
    FLocal := TDOMElement(FLocal.FirstChild);
end;

function TNXMLReader.EndWriting: Boolean;
begin
  Result := True;
end;

function TNXMLReader.BeginObject(var AClassName: string): Boolean;
begin
  AClassName := '';
  Result := False;
  if not CheckLocalType(sXObject) then Exit;
  Result := True;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXClass));
  FLocal := TDOMElement(FLocal.FirstChild);
  if FAttr <> nil then AClassName := FAttr.NodeValue;
end;

function TNXMLReader.EndObject: Boolean;
begin
  FLocal := TDOMElement(FLocal.ParentNode);
  SkipValue;
  Result := True;
end;

function TNXMLReader.BeginProperty(var Name: string): Boolean;
begin
  Name := '';
  Result := False;
  if FLocal.TagName <> sXProperty then Exit;
  Result := True;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXName));
  if FAttr <> nil then Name := FAttr.NodeValue;
end;

function TNXMLReader.EndProperty: Boolean;
begin
  Result := True;
end;

function TNXMLReader.BeginCustom: Boolean;
begin
  Result := False;
  if CheckLocalType(sXCustom) then
  begin
    Result := True;
    FLocal := TDOMElement(FLocal.FirstChild);
  end;
end;

function TNXMLReader.EndCustom: Boolean;
begin
  FLocal := TDOMElement(FLocal.ParentNode);
  SkipValue;
  Result := True;
end;

function TNXMLReader.IsBeginObject: Boolean;
begin
  Result := CheckLocalType(sXObject);
end;

function TNXMLReader.IsEndObject: Boolean;
begin
  Result := IsEndList;
end;

function TNXMLReader.IsEndList: Boolean;
begin
  Result := FLocal.NextSibling = nil;
end;

function TNXMLReader.IsEndCustom: Boolean;
begin
  Result := IsEndList;
end;

function TNXMLReader.ReadSet(PropInfo: PPropInfo): Long;
var
  S: string;
  EnumType: PTypeInfo;
  EnumName: string;
  P: Long;
begin
  if CheckLocalType(sXSet) then
  begin
    S := GetLocalStr;
    S := Copy(S, 2, Length(S) - 1);
    S := Copy(S, 1, Length(S) - 1);
    EnumType := GetTypeData(PropInfo^.PropType^)^.CompType^;
    Result := 0;
    while True do
    begin
      P := Pos(',', S);
      if P = 0 then P := Length(S) + 1;
      EnumName := Copy(S, 1, P - 1);
      Delete(S, 1, P);
      Include(TIntegerSet(Result), GetEnumValue(EnumType, EnumName));
      if S = '' then Break;
    end;
    SkipValue;
  end
  else
    Result := 0;
end;

function TNXMLReader.ReadMethod: string;
begin
  Result := ReadString;
end;

function TNXMLReader.CheckLocalType(const TypeStr: string): Boolean;
begin
  Result := False;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXType));
  if FAttr = nil then Exit;
  if FAttr.NodeValue = TypeStr then Result := True;
end;

function TNXMLReader.GetLocalStr: string;
begin
  Result := '';
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXValue));
  if FAttr <> nil then Result := FAttr.NodeValue;
end;

function TNXMLReader.GetLocalInt: Int64;
begin
  Result := 0;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXValue));
  if FAttr <> nil then Result := StrToInt64(FAttr.NodeValue);
end;

function TNXMLReader.GetLocalFloat: Double;
begin
  Result := 0.0;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXValue));
  if FAttr <> nil then Result := StrToFloat(FAttr.NodeValue);
end;

procedure TNXMLReader.GetLocalBinary(var Data: Pointer; var Count: Long);
var
  Buff: string;
begin
  Data := nil;
  Count := 0;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXValue));
  if FAttr = nil then Exit;
  Buff := FAttr.NodeValue;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXSize));
  if FAttr = nil then Exit;
  Count := StrToInt(FAttr.NodeValue);
  GetMem(Data, Count);
  Classes.HexToBin(PChar(Buff), Data, Count);
end;

procedure TNXMLReader.Flush;
begin
  try
    try
      FreeAndNil(FDoc);
      DOM.ReadXMLFile(FDoc, FStream);
    except
      on E: Exception do
        ShowError(ErrorStr(Self, 'Flush', E.Message), E.ClassName);
    end;
  finally
    if FDoc = nil then FDoc := TXMLDocument.Create;
  end;
end;

function TNXMLReader.BeginList(const Name: string): Boolean;
begin
  Result := False;
  FAttr := TDOMElement(FLocal.Attributes.GetNamedItem(sXType));
  if FAttr = nil then Exit;
  if FAttr.NodeValue = Name then
  begin
    Result := True;
    FLocal := TDOMElement(FLocal.FirstChild);
  end;
end;

function TNXMLReader.EndList: Boolean;
begin
  FLocal := TDOMElement(FLocal.ParentNode);
  SkipValue;
  Result := True;
end;

function TNXMLReader.ReadString: string;
begin
  if CheckLocalType(sXChar) or CheckLocalType(sXString)
    or CheckLocalType(sXIdent) or CheckLocalType(sXBool)
    or CheckLocalType(sXMethod) then
  begin
    Result := GetLocalStr;
    SkipValue;
  end
  else
    Result := '';
end;

function TNXMLReader.ReadIdent: string;
begin
  Result := ReadString;
end;

function TNXMLReader.ReadChar: Char;
begin
  Result := ReadString[1];
end;

function TNXMLReader.ReadLong: Long;
begin
  if CheckLocalType(sXLong) then
  begin
    Result := GetLocalInt;
    SkipValue;
  end
  else
    Result := 0;
end;

function TNXMLReader.ReadULong: ULong;
begin
  if CheckLocalType(sXULong) then
  begin
    Result := GetLocalInt;
    SkipValue;
  end
  else
    Result := 0;
end;

function TNXMLReader.ReadInt64: Int64;
begin
  if CheckLocalType(sXInt64) or CheckLocalType(sXULong)
    or CheckLocalType(sXLong) then
  begin
    Result := GetLocalInt;
    SkipValue;
  end
  else
    Result := 0;
end;

function TNXMLReader.ReadBool: Boolean;
begin
  if CheckLocalType(sXBool) then
  begin
    Result := SameStr(GetLocalStr, sTrue);
    SkipValue;
  end
  else
    Result := False;
end;

function TNXMLReader.ReadSingle: Single;
begin
  Result := ReadDouble;
end;

function TNXMLReader.ReadDouble: Double;
begin
  if CheckLocalType(sXSingle) or CheckLocalType(sXDouble) then
  begin
    Result := GetLocalFloat;
    SkipValue;
  end
  else
    Result := 0.0;
end;

procedure TNXMLReader.ReadBinary(var Data: Pointer; var Count: Long);
begin
  if CheckLocalType(sXBinary) then
  begin
    GetLocalBinary(Data, Count);
    SkipValue;
  end;
end;

procedure TNXMLReader.SkipValue;
begin
  if FLocal.NextSibling <> nil then
    FLocal := TDOMElement(FLocal.NextSibling);
end;


{*************************************************************************}
{                         TNXMLWriter methods                             }
{*************************************************************************}


constructor TNXMLWriter.Create(AStream: TNStream);
begin
  inherited Create(AStream);
  FDoc := TXMLDocument.Create;
  FDoc.WriteHeaders := False;
  FStack := TNStack.Create;
  FLocal := FDoc.CreateElement(sXMLHeader);
  FDoc.AppendChild(FLocal);
end;

destructor TNXMLWriter.Destroy;
begin
  FDoc.Free;
  FStack.Free;
  inherited Destroy;
end;

procedure TNXMLWriter.BeginWriting;
begin
end;

procedure TNXMLWriter.EndWriting;
begin
end;

procedure TNXMLWriter.BeginObject(Instance: TNPersistent);
begin
  FStack.Push(FLocal);
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXObject);
  FNode.SetAttribute(sXClass,
    GetTypeData(PTypeInfo(Instance.ClassType.ClassInfo)).UnitName + '.' +
    Instance.ClassName);
  FLocal := FNode;
end;

procedure TNXMLWriter.EndObject;
begin
  FLocal := TDOMElement(FStack.Pop);
end;

procedure TNXMLWriter.BeginProperty(const Name: string);
begin
  FPropNode := FDoc.CreateElement(sXProperty);
  FPropNode.SetAttribute(sXName, Name);
  FLocal.AppendChild(FPropNode);
end;

procedure TNXMLWriter.EndProperty;
begin
  FPropNode := nil;
end;

procedure TNXMLWriter.BeginCustom;
begin
  FStack.Push(FLocal);
  FPropNode.SetAttribute(sXType, sXCustom);
  FLocal := FPropNode;
  FPropNode := nil;
end;

procedure TNXMLWriter.EndCustom;
begin
  FLocal := TDOMElement(FStack.Pop);
end;

procedure TNXMLWriter.WriteSet(Value: Long; PropInfo: PPropInfo);
var
  I: Long;
  BaseType: PTypeInfo;
  S: string;
begin
  BaseType := GetTypeData(PropInfo^.PropType^)^.CompType^;
  S := '[';
  for I := 0 to SizeOf(TIntegerSet) * 8 - 1 do
    if I in TIntegerSet(Value) then
      S := S + GetEnumName(BaseType, I) + ',';
  if Length(S) > 1 then
    SetLength(S, Length(S) - 1);
  S := S + ']';
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXSet);
  FNode.SetAttribute(sXValue, S);
end;

procedure TNXMLWriter.WriteMethod(const Value: string);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXMethod);
  FNode.SetAttribute(sXValue, Value);
end;

function TNXMLWriter.GetNewNode(const Name: string; Global: Boolean):
  TDOMElement;
var
  TagName: string;
begin
  if Name = '' then
    TagName := sXVariable
  else
    TagName := Name;
  if ((FPropNode = nil) or (FLocal = FPropNode)) or Global then
  begin
    Result := FDoc.CreateElement(TagName);
    FLocal.AppendChild(Result);
  end
  else
    Result := FPropNode;
end;

procedure TNXMLWriter.Flush;
begin
  try
    DOM.WriteXMLFile(FDoc, FStream);
  except
    on E: Exception do ShowError(E.Message, E.ClassName);
  end;
end;

procedure TNXMLWriter.BeginList(const Name: string);
begin
  FStack.Push(FLocal);
  FNode := GetNewNode(sXList, True);
  FNode.SetAttribute(sXType, Name);
  FLocal := FNode;
end;

procedure TNXMLWriter.EndList;
begin
  FLocal := TDOMElement(FStack.Pop);
end;

procedure TNXMLWriter.WriteString(const Value: string);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXString);
  FNode.SetAttribute(sXValue, Value);
end;

procedure TNXMLWriter.WriteIdent(const Value: string);
begin
  if SameStr(Value, sFalse) then
    WriteBool(False)
  else
    if SameStr(Value, sTrue) then
      WriteBool(True)
    else
    begin
      FNode := GetNewNode;
      FNode.SetAttribute(sXType, sXIdent);
      FNode.SetAttribute(sXValue, Value);
    end;
end;

procedure TNXMLWriter.WriteChar(const Value: Char);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXChar);
  FNode.SetAttribute(sXValue, Value);
end;

procedure TNXMLWriter.WriteInt(const Value: Long);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXLong);
  FNode.SetAttribute(sXValue, IntToStr(Value));
end;

procedure TNXMLWriter.WriteInt(const Value: ULong);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXULong);
  FNode.SetAttribute(sXValue, IntToStr(Value));
end;

procedure TNXMLWriter.WriteInt(const Value: Int64);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXInt64);
  FNode.SetAttribute(sXValue, IntToStr(Value));
end;

procedure TNXMLWriter.WriteBool(const Value: Boolean);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXBool);
  FNode.SetAttribute(sXValue, Iff(Value, sTrue, sFalse));
end;

procedure TNXMLWriter.WriteFloat(const Value: Single);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXSingle);
  FNode.SetAttribute(sXValue, Format('%.7f', [Value]));
end;

procedure TNXMLWriter.WriteFloat(const Value: Double);
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXDouble);
  FNode.SetAttribute(sXValue, Format('%.14f', [Value]));
end;

procedure TNXMLWriter.WriteBinary(Data: Pointer; Count: Long);
var
  Text: PChar;
begin
  FNode := GetNewNode;
  FNode.SetAttribute(sXType, sXBinary);
  FNode.SetAttribute(sXSize, IntToStr(Count));
  Text := StrAlloc(Count * 2);
  BinToHex(Data, Text, Count);
  FNode.SetAttribute(sXValue, Text);
  StrDispose(Text);
end;

end.

